import pandas as pd

import sdmx


def pd_obj(specimen, name):
    """Return the specimen at *name* read, then converted to pandas."""
    with specimen(name) as f:
        data_msg = sdmx.read_sdmx(f)
    return sdmx.to_pandas(data_msg, rtype="compat")


def test_xml(specimen):
    """Test that objects have the expected layout with rtype='compat'."""
    # GenericData with non-time Dimension at observation → data frame
    df1 = pd_obj(specimen, "ng-xs.xml")
    assert isinstance(df1.index, pd.Index) and df1.index.name == "CURRENCY"
    assert isinstance(df1.columns, pd.MultiIndex)

    # TimeSeriesData → data frame with DatetimeIndex
    df2 = pd_obj(specimen, "ng-ts.xml")
    assert isinstance(df2.index, pd.DatetimeIndex)
    assert isinstance(df2.columns, pd.MultiIndex)

    # GenericData with AllDimensions at observation → series
    df3 = pd_obj(specimen, "ng-flat.xml")
    assert isinstance(df3, pd.Series)
    assert isinstance(df3.index, pd.MultiIndex)



import pytest

from sdmx.model import v21 as model
from sdmx.source import Source, add_source, get_source, list_sources, sources


def test_get_source(caplog):
    s1 = get_source("WB")
    assert 0 == len(caplog.messages)

    s2 = get_source("wb")
    assert "'WB' as a case-insensitive match for id 'wb'" in caplog.messages[-1]

    assert s1 == s2


def test_list_sources():
    source_ids = list_sources()
    # Correct number of sources, excluding those created for testing
    assert 34 == len(set(source_ids) - {"MOCK", "TEST"})

    # Listed alphabetically
    assert "ABS" == source_ids[0]
    assert "WB_WDI" == source_ids[-1]


def test_source_support():
    # Implicitly supported endpoint
    assert sources["ILO"].supports["categoryscheme"] is True

    # Specifically unsupported endpoint
    assert sources["ESTAT"].supports["contentconstraint"] is False

    # Explicitly supported structure-specific data
    assert sources["INEGI"].supports["structure-specific data"] is True


def test_add_source():
    profile = """{
        "id": "FOO",
        "name": "Demo source",
        "url": "https://example.org/sdmx"
        }"""
    add_source(profile)

    # JSON sources do not support metadata endpoints, by default
    profile2 = """{
        "id": "BAR",
        "data_content_type": "JSON",
        "name": "Demo source",
        "url": "https://example.org/sdmx"
        }"""
    add_source(profile2)
    assert not sources["BAR"].supports["datastructure"]

    with pytest.raises(
        ValueError, match="Data source 'ECB' already defined; use override=True"
    ):
        add_source(dict(id="ECB", name="Demo source", url="https://example.com/sdmx"))


class TestSource:
    @pytest.fixture
    def s(self):
        """An instance of the class."""
        yield Source(id="FOO", name="Test source", url="https://example.com")

    def test_get_url_class(self):
        """get_url_class() returns :class:`.v30.URL` as appropriate."""
        from sdmx.rest import v30

        assert issubclass(sources["ESTAT3"].get_url_class(), v30.URL)

    def test_modify_request_args(self, s):
        kwargs = dict(dsd=model.DataStructureDefinition())

        s.modify_request_args(kwargs)
        assert "structurespecificdata" in kwargs["headers"]["Accept"]



import pytest


class TestOfflineAdapter:
    def test_raises(self, session_with_stored_responses):
        with pytest.raises(RuntimeError):
            session_with_stored_responses.get("https://example.com")



import io
import logging
from datetime import datetime

import pytest
from lxml import etree

import sdmx
import sdmx.writer.xml
from sdmx import message
from sdmx.format.xml import validate_xml
from sdmx.model import common, v21
from sdmx.model import v21 as m
from sdmx.model.v21 import DataSet, DataStructureDefinition, Dimension, Key, Observation
from sdmx.writer.xml import writer as XMLWriter

log = logging.getLogger(__name__)

# Fixtures


@pytest.fixture
def structure_message() -> message.StructureMessage:
    """A StructureMessage that serializes to XSD-valid SDMX-XML."""
    return message.StructureMessage(
        header=message.Header(
            id="N_A",
            prepared=datetime.now(),
            sender=common.Agency(id="N_A"),
        )
    )


@pytest.fixture
def dsd():
    dsd = DataStructureDefinition()

    for order, id in enumerate(["FOO", "BAR", "BAZ"]):
        dsd.dimensions.append(Dimension(id=id, order=order))

    return dsd


@pytest.fixture
def obs(dsd):
    return Observation(dimension=dsd.make_key(Key, dict(FOO=1, BAR=2)), value=42.0)


@pytest.fixture
def dks(dsd):
    dim = dsd.dimensions.get("FOO")
    yield m.DataKeySet(
        included=True,
        keys=[
            m.DataKey(
                included=True,
                key_value={dim: m.ComponentValue(value_for=dim, value="foo0")},
            )
        ],
    )


# Test specific methods associated with specific classes


class TestNameableArtefact:
    def test_xsd(self, structure_message):
        """Annotations for a NameableArtefact are output in the correct order.

        In https://github.com/khaeru/sdmx/issues/210 it was reported that
        <com:Annotations> incorrectly appeared before <com:Name>.
        """
        # Common arguments
        args = dict(
            # Identifiable Artefact
            id="FOO",
            # Nameable Artefact
            name="foo",
            description="bar",
            # VersionableArtefact
            version="1",
            # MaintainableArtefact
            maintainer=common.Agency(id="N_A"),
            is_external_reference=False,
            is_final=True,
        )
        dsd = v21.DataStructureDefinition(**args)
        na = v21.DataflowDefinition(
            annotations=[common.Annotation(id="baz", text="qux")],  # Annotable Artefact
            **args,  # Identifiable, Nameable, Versionable, Maintainable
            structure=dsd,  # Dataflow-specific attributes
        )
        structure_message.dataflow[na.id] = na

        # Write to SDMX-ML
        buf = io.BytesIO(sdmx.to_xml(structure_message))

        # Validate using XSD. Fails with v2.19.1.
        assert validate_xml(buf), buf.getvalue().decode()


def test_contact() -> None:
    c = m.Contact(
        name="John Smith",
        org_unit="Human Resources",
        telephone="+1234567890",
        uri=["https://example.org"],
        email=["john.smith@example.org"],
    )

    result = sdmx.to_xml(c, pretty_print=True)

    assert result.decode().endswith(
        """
  <com:Name xml:lang="en">John Smith</com:Name>
  <str:Department xml:lang="en">Human Resources</str:Department>
  <str:Telephone>+1234567890</str:Telephone>
  <str:URI>https://example.org</str:URI>
  <str:Email>john.smith@example.org</str:Email>
</str:Contact>
"""
    )


def test_codelist(codelist):
    result = sdmx.to_xml(codelist, pretty_print=True)
    print(result.decode())


def test_DataKeySet(dks):
    """:class:`.DataKeySet` can be written to XML."""
    sdmx.to_xml(dks)


def test_ContentConstraint(dsd, dks):
    """:class:`.ContentConstraint` can be written to XML."""
    sdmx.to_xml(
        m.ContentConstraint(
            role=m.ConstraintRole(role=m.ConstraintRoleType.allowable),
            content=[dsd],
            data_content_keys=dks,
        )
    )


def test_ds(dsd, obs) -> None:
    # Write DataSet with Observations not in Series
    ds = DataSet(structured_by=dsd)
    ds.obs.append(obs)

    result = sdmx.to_xml(ds, pretty_print=True)
    # print(result.decode())
    del result


def test_ds_structurespecific(dsd):
    series_key = dsd.make_key(m.SeriesKey, dict(FOO=1, BAR=2))
    primary_measure = m.PrimaryMeasure(id="OBS_VALUE")
    observations = [
        m.Observation(
            series_key=series_key,
            dimension=dsd.make_key(m.Key, dict(BAZ=3)),
            value_for=primary_measure,
            value=25,
        ),
        m.Observation(
            series_key=series_key,
            dimension=dsd.make_key(m.Key, dict(BAZ=4)),
            value_for=primary_measure,
            value=0,
        ),
    ]
    series = {series_key: observations}
    ds = m.StructureSpecificDataSet(structured_by=dsd, series=series)
    dm = message.DataMessage(data=[ds])
    result = sdmx.to_xml(dm, pretty_print=True)
    exp = (
        '    <Series FOO="1" BAR="2">\n'
        '      <Obs OBS_VALUE="25" BAZ="3"/>\n'
        '      <Obs OBS_VALUE="0" BAZ="4"/>\n'
        "    </Series>"
    )
    assert exp in result.decode()


def test_obs(obs):
    # Generate <gen:ObsKey> element for 2+-dimensional Observation.dimension
    exp = (
        '<gen:ObsKey><gen:Value id="FOO" value="1"/>'
        '<gen:Value id="BAR" value="2"/></gen:ObsKey>'
    )
    assert exp in sdmx.to_xml(obs).decode()

    # Exception raised in structure-specific data because `obs` fixture has no value_for
    with pytest.raises(
        ValueError,
        match="Observation.value_for is None when writing structure-specific data",
    ):
        XMLWriter.recurse(obs, struct_spec=True)


def test_reference() -> None:
    cl = m.Codelist(id="FOO", version="1.0")
    c = m.Code(id="BAR")
    cl.append(c)

    # <Ref …> to Item has maintainableParentVersion, but no version
    result = sdmx.writer.xml.reference(c, style="Ref")
    result_str = etree.tostring(result).decode()
    assert 'maintainableParentVersion="1.0"' in result_str
    assert 'version="1.0"' not in result_str

    # <Ref …> to ItemScheme has version, but not maintainableParentVersion
    result = sdmx.writer.xml.reference(cl, style="Ref")
    result_str = etree.tostring(result).decode()
    assert 'maintainableParentVersion="1.0"' not in result_str
    assert 'version="1.0"' in result_str


def test_VersionableArtefact() -> None:
    """:class:`VersionableArtefact` with :class:`.Version` instance can be written."""
    cl: common.Codelist = common.Codelist(id="FOO", version=common.Version("1.2.3"))

    # Written to XML without error
    result = sdmx.to_xml(cl).decode()
    assert 'version="1.2.3"' in result


# sdmx.message classes


def test_Footer(footer):
    """:class:`.Footer` can be written."""
    sdmx.to_xml(footer)


def test_structuremessage(tmp_path, structuremessage):
    result = sdmx.to_xml(structuremessage, pretty_print=True)

    # Message can be round-tripped to/from file
    path = tmp_path / "output.xml"
    path.write_bytes(result)
    msg = sdmx.read_sdmx(path)

    # Contents match the original object
    assert (
        msg.codelist["CL_COLLECTION"]["A"].name["en"]
        == structuremessage.codelist["CL_COLLECTION"]["A"].name["en"]
    )

    # False because `structuremessage` lacks URNs, which are constructed automatically
    # by `to_xml`
    assert not msg.compare(structuremessage, strict=True)
    # Compares equal when allowing this difference
    assert msg.compare(structuremessage, strict=False)


def test_DataMessage(datamessage):
    """:class:`.DataMessage` can be written."""
    sdmx.to_xml(datamessage)


def test_ErrorMessage(errormessage):
    """:class:`.ErrorMessage` can be written."""
    sdmx.to_xml(errormessage)


@pytest.mark.parametrize(
    "data_id, structure_id",
    [
        (
            "INSEE/CNA-2010-CONSO-SI-A17.xml",
            "INSEE/CNA-2010-CONSO-SI-A17-structure.xml",
        ),
        ("INSEE/IPI-2010-A21.xml", "INSEE/IPI-2010-A21-structure.xml"),
        ("ECB_EXR/1/M.USD.EUR.SP00.A.xml", "ECB_EXR/1/structure.xml"),
        ("ECB_EXR/ng-ts.xml", "ECB_EXR/ng-structure-full.xml"),
        ("ECB_EXR/ng-ts-ss.xml", "ECB_EXR/ng-structure-full.xml"),
        ("ESTAT/esms.xml", "ESTAT/esms-structure.xml"),
        # DSD reference does not round-trip correctly
        pytest.param(
            "ECB_EXR/rg-xs.xml",
            "ECB_EXR/rg-structure-full.xml",
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        # Example of a not-implemented feature: DataSet with groups
        pytest.param(
            "ECB_EXR/sg-ts-gf.xml",
            "ECB_EXR/sg-structure-full.xml",
            marks=pytest.mark.xfail(raises=NotImplementedError),
        ),
    ],
)
def test_data_roundtrip(pytestconfig, specimen, data_id, structure_id, tmp_path):
    """Test that SDMX-ML DataMessages can be 'round-tripped'."""

    # Read structure from file
    with specimen(structure_id) as f:
        dsd = sdmx.read_sdmx(f).structure[0]

    # Read data from file, using the DSD
    with specimen(data_id) as f:
        msg0 = sdmx.read_sdmx(f, structure=dsd)

    # Write to file
    path = tmp_path / "output.xml"
    path.write_bytes(sdmx.to_xml(msg0, pretty_print=True))

    # Read again, using the same DSD
    msg1 = sdmx.read_sdmx(path, structure=dsd)

    # Contents are identical
    assert msg0.compare(msg1, strict=True), (
        path.read_text() if pytestconfig.getoption("verbose") else path
    )


@pytest.mark.parametrize(
    "specimen_id, strict",
    [
        ("BIS/gh-180.xml", False),
        ("ECB/orgscheme.xml", True),
        ("ECB_EXR/1/structure-full.xml", False),
        ("ESTAT/apro_mk_cola-structure.xml", True),
        ("ESTAT/esms-structure.xml", True),
        pytest.param(
            "ISTAT/47_850-structure.xml", True, marks=[pytest.mark.skip(reason="Slow")]
        ),
        ("IMF/ECOFIN_DSD-structure.xml", True),
        ("INSEE/CNA-2010-CONSO-SI-A17-structure.xml", False),
        ("INSEE/IPI-2010-A21-structure.xml", False),
        ("INSEE/dataflow.xml", False),
        ("OECD/actualconstraint-0.xml", True),
        ("SGR/common-structure.xml", True),
        ("UNSD/codelist_partial.xml", True),
        ("TEST/gh-149.xml", False),
    ],
)
def test_structure_roundtrip(specimen, specimen_id, strict, tmp_path):
    """Test that SDMX-ML StructureMessages can be 'round-tripped'."""

    # Read a specimen file
    with specimen(specimen_id) as f:
        msg0 = sdmx.read_sdmx(f)

    # Write to a bytes buffer
    data = io.BytesIO(sdmx.to_xml(msg0, pretty_print=True))

    # Read again
    msg1 = sdmx.read_sdmx(data)

    # Contents are identical
    try:
        assert msg0.compare(msg1, strict)
    except AssertionError:  # pragma: no cover
        path = tmp_path.joinpath("output.xml")
        path.write_bytes(data.getbuffer())
        log.error(f"compare() = False; see {path}")
        raise



import csv
import logging
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from itertools import zip_longest
from typing import TYPE_CHECKING, Literal, MutableSequence, Optional, Sequence, Union

import sdmx.message
from sdmx.format import list_media_types
from sdmx.model import common, v30
from sdmx.reader import base
from sdmx.reader.base import BaseReader

if TYPE_CHECKING:
    from typing import TypedDict

    import pandas

    from sdmx.model import v21

    class DataSetKwargs(TypedDict):
        described_by: Optional[common.BaseDataflow]
        structured_by: common.BaseDataStructureDefinition


log = logging.getLogger(__name__)


@dataclass
class Options:
    """SDMX-CSV 2.0.0 format options."""

    #: Types of labels included. Appears in the specification.
    labels: Literal["both", "id", "name"] = "id"

    #: Whether series, observation, or no keys are expressed in their own columns (in
    #: addition to dimension columns). Appears in the specification.
    key: Literal["both", "none", "obs", "series"] = "none"

    #: “Custom columns” detected by :meth:`.Reader.inspect_header`.
    custom_columns: list[bytes] = field(default_factory=list)

    #: CSV field delimiter.
    delimiter: str = ","

    #: SDMX-CSV “sub-field” delimiter.
    delimiter_sub: str = ""


class Reader(BaseReader):
    """Read SDMX-CSV."""

    # BaseReader attributes
    media_types = list_media_types(base="csv")
    suffixes = [".csv"]

    #: Handlers for individual fields in a CSV record. This collection has exactly the
    #: same number of handlers as columns in the `data` passed to :meth:`convert`.
    handlers: Sequence["Handler"]

    _dataflow: Optional["common.BaseDataflow"]
    _structure: Union["v21.DataStructureDefinition", "v30.DataStructure"]
    _observations: dict[tuple[str, str, str], list["common.BaseObservation"]]

    def __init__(self):
        self.options = Options()
        self.handlers = []
        self._dataflow = None
        self._structure = None
        self._observations = defaultdict(list)

    def convert(self, data, structure=None, *, delimiter: str = ",", **kwargs):
        """Read a message from `data`."""
        self.options.delimiter = delimiter

        if isinstance(structure, common.BaseDataflow):
            self._dataflow = structure
            self._structure = structure.structure
        else:
            self._dataflow = None
            self._structure = structure

        # Create a CSV reader
        lines = data.read().decode().splitlines()
        reader = csv.reader(lines, delimiter=self.options.delimiter)

        self.inspect_header(next(reader))

        # Parse remaining rows to observations
        for i, row in enumerate(reader):
            self.handle_row(row)

        # Create a data message
        self.message = sdmx.message.DataMessage(dataflow=self._dataflow)

        # Create 1 data set for each of the 4 ActionType values
        ds_kw: "DataSetKwargs" = dict(
            described_by=self._dataflow, structured_by=self._structure
        )
        for (*_, action), obs in self._observations.items():
            a = common.ActionType[
                {"A": "append", "D": "delete", "I": "information", "R": "replace"}[
                    action
                ]
            ]

            self.message.data.append(v30.DataSet(action=a, **ds_kw))
            self.message.data[-1].add_obs(obs)

        return self.message

    def handle_row(self, row: list[str]) -> None:
        """Handle a single CSV row."""
        obs = v30.Observation(
            dimension=v30.Key(),
            attached_attribute={"__TARGET": v30.AttributeValue(value=[])},
        )

        for h, v in zip_longest(self.handlers, row):
            h(obs, v)

        # Remove the "__TARGET" annotation and construct a key describing the data set
        # to which this observation belongs
        target = tuple(obs.attached_attribute.pop("__TARGET").value)

        self._observations[target].append(obs)

    def inspect_header(self, header: list[str]) -> None:  # noqa: C901  TODO Reduce complexity from 12 → ≤10
        """Inspect the SDMX-CSV header and determine the format :class:`.Options`.

        Raises
        ------
        ValueError
            if the data contain malformed SDMX-CSV 2.0.0.
        """
        handlers: MutableSequence[Optional["Handler"]] = [
            StoreTarget(allowable={"dataflow", "dataprovision", "datastructure"}),
            StoreTarget(),
        ] + ([None] * (len(header) - 2))

        # Columns in fixed order

        if match := re.fullmatch(r"STRUCTURE(\[(?P<delimiter_sub>.)\])?", header[0]):
            self.options.delimiter_sub = match.groupdict().get("delimeter_sub", None)
        else:
            raise ValueError(
                f"Invalid SDMX-CSV 2.0.0: {header[0]!r} in line 1, field 1; "
                "expected 'STRUCTURE' or 'STRUCTURE[…]'"
            )

        if not header[1] == "STRUCTURE_ID":
            raise ValueError(
                f"Invalid SDMX-CSV 2.0.0: {header[1]!r} in line 1, field 2; "
                "expected 'STRUCTURE_ID'"
            )

        i = 2
        if header[i] == "STRUCTURE_NAME":
            self.options.labels = "name"
            handlers[i] = Name()
            i += 1

        # Maybe a column "ACTION"
        if header[i] == "ACTION":
            handlers[i] = StoreTarget(allowable=set("ADIR"))
            i += 1

        if i < len(header) and header[i] == "SERIES_KEY":
            self.options.key = "series"
            handlers[i] = SeriesKey()
            i += 1

        if i < len(header) and header[i] == "OBS_KEY":
            handlers[i] = ObsKey()
            self.options.key = {"none": "obs", "series": "both"}.get(self.options.key)
            i += 1

        # From this point, columns may appear in any order

        inspected = set(range(i))

        for cls, components, multi_possible in (
            (KeyValue, self._structure.dimensions, False),
            (ObsValue, self._structure.measures, False),
            (AttributeValue, self._structure.attributes, True),
        ):
            for c in components:
                pattern = re.compile(
                    c.id + (r"(?P<multi>\[\])?" if multi_possible else "") + "(|: .*)"
                )
                matches = list(filter(None, map(pattern.fullmatch, header[i:])))
                if not len(matches):
                    log.warning(f"No column detected for {c!r}")
                    continue

                idx = header.index(matches[0].string)
                handlers[idx] = cls(c, multi="multi" in matches[0].groupdict())
                inspected.add(idx)

                if self.options.labels == "name":
                    handlers[idx + 1] = Name()
                    inspected.add(idx + 1)

        for i in set(range(len(header))) - inspected:
            h = header[i]
            handlers[i] = Custom(h)
            self.options.custom_columns.append(h)

        self.handlers = tuple(filter(None, handlers))
        assert len(self.handlers) == len(header)


class DataFrameConverter(base.Converter):
    @classmethod
    def handles(cls, data, kwargs) -> bool:
        import pandas as pd

        return isinstance(data, pd.DataFrame) and "structure" in kwargs

    def convert(
        self, data: "pandas.DataFrame", structure=None, **kwargs
    ) -> "sdmx.message.DataMessage":
        assert 0 == len(kwargs)

        # TEMPORARY Use a Reader instance
        r = Reader()
        r._dataflow = structure
        r._structure = structure.structure
        r.inspect_header(data.columns.to_list())

        # Parse remaining rows to observations
        for _, row in data.iterrows():
            r.handle_row(row.to_list())

        # Create a data message
        message = sdmx.message.DataMessage(dataflow=r._dataflow)

        # Create 1 data set for each of the 4 ActionType values
        ds_kw: "DataSetKwargs" = dict(
            described_by=r._dataflow, structured_by=r._structure
        )
        for (*_, action), obs in r._observations.items():
            a = common.ActionType[
                {"A": "append", "D": "delete", "I": "information", "R": "replace"}[
                    action
                ]
            ]

            message.data.append(v30.DataSet(action=a, **ds_kw))
            message.data[-1].add_obs(obs)

        return message


class Handler(ABC):
    """Base class for :attr:`.Reader.handlers`."""

    @abstractmethod
    def __call__(self, obs: "common.BaseObservation", value: str) -> None:
        """Handle the `value` in one field/record, and update the resulting `obs`.

        Subclasses **must** implement this method.
        """

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"


class Name(Handler):
    """Handler for :py:`Options.labels == "name"` columns.

    Does nothing; the values are not stored.
    """

    def __call__(self, obs, value):
        pass


class NotHandled(Handler):
    """Handler that does nothing."""

    def __call__(self, obs, value):
        log.info(f"Not handled: {self.__class__.__name__}: {value}")


class StoreTarget(Handler):
    """Store `value` on a special :class:`.DataAttribute` with ID "__TARGET" on `obs`.

    Used for the STRUCTURE, STRUCTURE_ID, and ACTION columns.
    """

    def __init__(self, allowable: Optional[set[str]] = None):
        self.allowable = allowable

    def __call__(self, obs, value):
        assert value in self.allowable if self.allowable else True
        obs.attached_attribute["__TARGET"].value.append(value)


class SeriesKey(NotHandled):
    """ "SERIES_KEY" columns are currently not handled."""

    pass


class ObsKey(NotHandled):
    """ "OBS_KEY" columns are currently not handled."""

    pass


class KeyValue(Handler):
    """Handle a :class:`~.common.KeyValue` for one :class:`.Dimension`."""

    def __init__(self, dimension, **kwargs):
        self.dimension = dimension

    def __call__(self, obs, value):
        obs.dimension.values[self.dimension.id] = v30.KeyValue(
            id=self.dimension.id, value=value, value_for=self.dimension
        )


class ObsValue(Handler):
    """Handle the :attr:`Observation.value <.BaseObservation.value>.

    In line with :mod:`.model.v30`, multiple values (for data structures with multiple
    measures) are currently not handled.
    """

    def __init__(self, measure, **kwargs):
        self.measure = measure

    def __call__(self, obs, value):
        obs.value = value


class AttributeValue(Handler):
    """Handle a :class:`.v30.AttributeValue` for one :class:`.DataAttribute`.

    The attribute value is attached to the `obs`.
    """

    def __init__(self, attribute, multi: bool):
        self.attribute = attribute
        if multi:
            log.info(f"Column {attribute.id!r}: multiple values will not be unpacked")

    def __call__(self, obs, value):
        obs.attached_attribute[self.attribute.id] = v30.AttributeValue(
            value=value, value_for=self.attribute
        )


class Custom(Handler):
    """Handler for custom columns.

    Currently values are ignored.

    .. todo:: Store as :class:`.Annotation` or temporary attribute values on `obs`.
    """

    def __init__(self, header: str):
        log.info(f"Column {header!r} detected as custom and will not be stored")

    def __call__(self, obs, value):
        pass



import operator
import re
from typing import Callable, Optional, Union

import packaging.version

#: Regular expressions (:class:`re.Pattern`) for version strings.
#:
#: - :py:`"2_1"` SDMX 2.1, e.g. "1.0"
#: - :py:`"3_0"` SDMX 3.0, e.g. "1.0.0-draft"
#: - :py:`"py"` Python-compatible versions, using :mod:`packaging.version`.
VERSION_PATTERNS = {
    "2_1": re.compile(r"^(?P<release>[0-9]+(?:\.[0-9]+){1})$"),
    "3_0": re.compile(r"^(?P<release>[0-9]+(?:\.[0-9]+){2})(-(?P<ext>.+))?$"),
    "py": re.compile(
        r"^\s*" + packaging.version.VERSION_PATTERN + r"\s*$",
        re.VERBOSE | re.IGNORECASE,
    ),
}


def _cmp_method(op: Callable) -> Callable:
    def cmp(self, other) -> bool:
        try:
            return op(self._key, other._key)
        except AttributeError:
            if isinstance(other, str):
                return op(self, Version(other))
            else:
                return NotImplemented

    return cmp


class Version(packaging.version.Version):
    """Class representing a version.

    The SDMX Information Model **does not** specify a Version class; instead,
    :attr:`.VersionableArtefact.version` is described as “a version **string** following
    SDMX versioning rules.”

    In order to simplify application of those ‘rules’, and to handle the differences
    between SDMX 2.1 and 3.0.0, this class extends :class:`packaging.version.Version`,
    which provides a complete interface for interacting with Python version specifiers.
    The extensions implement the particular form of versioning laid out by the SDMX
    standards. Specifically:

    - :attr:`kind` as added to identify whether a Version instance is an SDMX 2.1, SDMX
      3.0, or Python-style version string.
    - Attribute aliases for particular terms used in the SDMX 3.0 standards:
      :attr:`patch` and :attr:`ext`.
    - The :class:`str` representation of a Version uses the SDMX 3.0 style of separating
      the :attr:`ext` with a hyphen ("1.0.0-dev1"), which differs from the Python style
      of using no separator for a ‘post-release’ ("1.0.0dev1") or a plus symbol for a
      ‘local part’ ("1.0.0+dev1").
    - :meth:`increment`, an added convenience method.
    - The class is comparable and interchangeable with :class:`str` version expressions.

    Parameters
    ----------
    version : str
        String expression
    """

    #: Type of version expression; one of the keys of :data:`.VERSION_PATTERNS`.
    kind: str

    def __init__(self, version: str):
        for kind, pattern in VERSION_PATTERNS.items():
            match = pattern.fullmatch(version)
            if match:
                break

        if not match:
            raise packaging.version.InvalidVersion(version)

        self.kind = kind

        if kind == "py":
            tmp = packaging.version.Version(version)
            self._version = tmp._version
        else:
            # Store the parsed out pieces of the version
            try:
                ext = match.group("ext")
                local = None if ext is None else (ext,)
            except IndexError:
                local = None
            self._version = packaging.version._Version(
                epoch=0,
                release=tuple(int(i) for i in match.group("release").split(".")),
                pre=None,
                post=None,
                dev=None,
                local=local,
            )

        self._update_key()

    def _update_key(self):
        # Generate a key which will be used for sorting
        self._key = packaging.version._cmpkey(
            self._version.epoch,
            self._version.release,
            self._version.pre,
            self._version.post,
            self._version.dev,
            self._version.local,
        )

    def __str__(self):
        if self.kind == "3_0":
            parts = [".".join(str(x) for x in self.release)]
            if self.ext:
                parts.append(f"-{self.ext}")
            return "".join(parts)
        else:
            return super().__str__()

    __eq__ = _cmp_method(operator.eq)
    __ge__ = _cmp_method(operator.ge)
    __gt__ = _cmp_method(operator.gt)
    __le__ = _cmp_method(operator.le)
    __lt__ = _cmp_method(operator.lt)
    __ne__ = _cmp_method(operator.ne)

    @property
    def patch(self) -> int:
        """Alias for :any:`Version.micro <packaging.version.Version.micro>`."""
        return self.micro

    @property
    def ext(self) -> Optional[str]:
        """SDMX 3.0 version 'extension'.

        For :py:`kind="py"`, this is equivalent to :attr:`Version.local
        <packaging.version.Version.local>`.
        """
        if self._version.local is None:
            return None
        else:
            return "".join(map(str, self._version.local))

    def increment(self, **kwargs: Union[bool, int]) -> "Version":
        """Return a Version that is incrementally greater than the current Version.

        If no arguments are given, then by default :py:`minor=True` and :py:`ext=1`.

        Parameters
        ----------
        major : bool or int, optional
            If given, increment the :attr:`Version.major
            <packaging.version.Version.major>` part.
        minor : bool or int, optional
            If given, increment the :attr:`Version.minor
            <packaging.version.Version.minor>` part.
        patch : bool or int, optional
            If given, increment the :attr:`.Version.patch` part.
        micro : bool or int, optional
            Alias for `patch`.
        ext : bool or int, optional
            If given, increment the :attr:`.Version.ext` part. If this part is not
            present, add "dev1".
        local: bool or int, optional
            Alias for `ext`.
        """
        if not kwargs:
            # Apply defaults
            kwargs["minor"] = kwargs["ext"] = 1

        # Convert self._version.release into a mutable dict
        N_release = len(self._version.release)  # Number of parts in `release` tuple
        parts = dict(
            major=self._version.release[0] if N_release > 0 else 0,
            minor=self._version.release[1] if N_release > 1 else 0,
            patch=self._version.release[2] if N_release > 2 else 0,
        )
        # Convert self._version.local into a mutable list
        local = list(self._version.local) if self._version.local is not None else []

        # Increment parts according to kwargs
        for part, value in kwargs.items():
            # Recognize kwarg aliases
            part = {"local": "ext", "micro": "patch"}.get(part, part)

            # Update the extension/local part
            if part == "ext":
                if not len(local):
                    ext = "dev1"
                elif match := re.fullmatch("([^0-9]+)([0-9]+)", str(local[0])):
                    _l, _n = match.group(1, 2)
                    ext = f"{_l}{int(_n) + value}"
                else:
                    raise NotImplementedError(
                        f"Increment SDMX version extension {self.ext!r}"
                    )
                local = [ext]
                continue

            try:
                # Update the major/minor/patch parts
                parts[part] += int(value)
            except KeyError:
                raise ValueError(f"increment(..., {part}={value})")

        # Construct a new Version object
        result = type(self)(str(self))
        # Overwrite its private _version attribute and key
        result._version = packaging.version._Version(
            epoch=self._version.epoch,
            release=tuple(parts.values()),
            pre=self._version.pre,
            post=self._version.post,
            dev=self._version.dev,
            local=tuple(local) if len(local) else None,
        )
        result._update_key()

        return result


def increment(value: Union[packaging.version.Version, str], **kwargs) -> Version:
    """Increment the version `existing`.

    Identical to :py:`Version(str(value)).increment(**kwargs)`.

    See also
    --------
    Version.increment
    """
    return Version(str(value)).increment(**kwargs)


def parse(value: str) -> Version:
    """Parse the given version string.

    Identical to :py:`Version(value)`.

    See also
    --------
    Version
    """
    return Version(value)



import logging
import re
import zipfile
from functools import lru_cache
from itertools import chain
from operator import itemgetter
from pathlib import Path
from shutil import copytree
from typing import IO, Iterable, Mapping, Optional, Union, cast

from lxml import etree
from lxml.etree import QName

from sdmx.format import Version

log = logging.getLogger(__name__)

# Tags common to SDMX-ML 2.1 and 3.0

# XML tag name ("str" namespace) and class name are the same
CT1 = [
    "Agency",
    "AgencyScheme",
    "Categorisation",
    "Category",
    "CategoryScheme",
    "Code",
    "Codelist",
    "Concept",
    "ConceptScheme",
    "CustomType",
    "CustomTypeScheme",
    "DataConsumer",
    "DataConsumerScheme",
    "DataProvider",
    "DataProviderScheme",
    "HierarchicalCode",
    "Level",
    "NamePersonalisation",
    "NamePersonalisationScheme",
    "Ruleset",
    "RulesetScheme",
    "TimeDimension",
    "TransformationScheme",
    "UserDefinedOperatorScheme",
]

# XML tag name and class name differ
CT2 = [
    ("model.Annotation", "com:Annotation"),
    ("model.Agency", "str:Agency"),  # Order matters
    ("model.Agency", "mes:Receiver"),
    ("model.Agency", "mes:Sender"),
    ("model.AttributeDescriptor", "str:AttributeList"),
    ("model.Concept", "str:ConceptIdentity"),
    ("model.Codelist", "str:Enumeration"),  # This could possibly be ItemScheme
    ("model.Dimension", "str:Dimension"),  # Order matters
    ("model.Dimension", "str:DimensionReference"),
    ("model.Dimension", "str:GroupDimension"),
    ("model.DataAttribute", "str:Attribute"),
    ("model.DataStructureDefinition", "str:DataStructure"),
    ("model.DimensionDescriptor", "str:DimensionList"),
    ("model.GroupDimensionDescriptor", "str:Group"),
    ("model.GroupDimensionDescriptor", "str:AttachmentGroup"),
    ("model.GroupKey", "gen:GroupKey"),
    ("model.Key", "gen:ObsKey"),
    ("model.MeasureDescriptor", "str:MeasureList"),
    ("model.MetadataStructureDefinition", "str:MetadataStructure"),
    ("model.SeriesKey", "gen:SeriesKey"),
    ("model.Structure", "com:Structure"),
    ("model.Structure", "str:Structure"),
    ("model.StructureUsage", "com:StructureUsage"),
    ("model.VTLMappingScheme", "str:VtlMappingScheme"),
    # Message classes
    ("message.DataMessage", "mes:StructureSpecificData"),
    ("message.MetadataMessage", "mes:GenericMetadata"),
    ("message.MetadataMessage", "mes:StructureSpecificMetadata"),
    ("message.ErrorMessage", "mes:Error"),
    ("message.StructureMessage", "mes:Structure"),
]

NS = {
    "": None,
    "xml": "http://www.w3.org/XML/1998/namespace",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    # To be formatted
    "com": "{}/common",
    "md": "{}/metadata/generic",
    "md_ss": "{}/metadata/structurespecific",
    "data": "{}/data/structurespecific",
    "str": "{}/structure",
    "mes": "{}/message",
    "gen": "{}/data/generic",
    "footer": "{}/message/footer",
}


def validate_xml(
    msg: Union[Path, IO],
    schema_dir: Optional[Path] = None,
    version: Union[str, Version] = Version["2.1"],
    max_errors: int = -1,
) -> bool:
    """Validate SDMX-ML in `msg` against the XML Schema (XSD) documents.

    A log message with level :data:`logging.ERROR` is emitted if validation fails. This
    indicates the first (possibly not only) element in `msg` that is not valid per the
    schemas.

    Parameters
    ----------
    msg
        Path or io-like containing an SDMX-ML message.
    schema_dir
        Directory with SDMX-ML XSD schemas used to validate the message.
    version
        The SDMX-ML schema version to validate against. One of ``2.1`` or ``3.0``.
    max_errors
        Maximum number of messages to log on validation failure.

    Returns
    -------
    bool
        :any:`True` if validation passed, otherwise :any:`False`.

    Raises
    ------
    FileNotFoundError
        if `schema_dir` (or a subdirectory) does not contain :file:`SDMXMessage.xsd`.
        Use :func:`sdmx.install_schemas` to download the schema files.
    NotImplementedError
        if `msg` contains valid XML, but with a root element that is not part of the
        SDMX-ML standard.
    """
    schema_dir, version = _handle_validate_args(schema_dir, version)

    # Find SDMXMessage.xsd in `schema_dir` or a subdirectory
    for candidate in schema_dir, schema_dir.joinpath(version.name):
        try:
            # Turn the XSD into a schema object
            xml_schema = etree.XMLSchema(file=candidate.joinpath("SDMXMessage.xsd"))
            break
        except Exception:
            xml_schema = None

    if xml_schema is None:
        raise FileNotFoundError(f"Could not find XSD files in {schema_dir}")

    # Parse the given document
    msg_doc = etree.parse(msg)

    if not xml_schema.validate(msg_doc):
        for i, entry in enumerate(
            cast(Iterable["etree._LogEntry"], xml_schema.error_log)
        ):
            if (
                i == 0
                and "No matching global declaration available for the validation root"
                in entry.message
            ):
                raise NotImplementedError(
                    f"Validate non-SDMX root element <{msg_doc.getroot().tag}>"
                ) from None
            elif i == max_errors:
                break
            log.log(getattr(logging, entry.level_name), entry.message)

        return False
    else:
        return True


def _extracted_zipball(version: Version, force: bool = False) -> Path:
    """Retrieve, cache, and extract the SDMX-ML schemas for `version`.

    1. Query the GitHub REST API to identify a URL for the `version` in zipball format.
    2. Download and cache the zipball. The file is not downloaded if it already exists.
    3. Unpack the archive.

    Actions (2) and (3) are performed in the user's cache directory (for instance,
    :file:`$HOME/.cache/sdmx/`). :func:`install_schemas` handles copying the extracted
    files to other locations.

    Returns
    -------
    Path
        Path to the root folder of the unpacked archive.
    """
    import platformdirs
    import requests

    # Map SDMX-ML schema versions to repo paths
    version_path = {Version["2.1"]: "v2.1", Version["3.0.0"]: "v3.0.0"}[version]

    # Check the latest release to get the URL to the schema zip
    url_base = "https://api.github.com/repos/sdmx-twg/sdmx-ml"
    headers = {
        "Accept": "application/vnd.github+json",
        "X-GitHub-Api-Version": "2022-11-28",
    }
    release_json = requests.get(
        url=f"{url_base}/releases/tags/{version_path}", headers=headers
    ).json()
    try:
        zipball_url = release_json["zipball_url"]
    except KeyError:  # pragma: no cover
        zipball_url = f"{url_base}/zipball/{version_path}"
        log.debug(f"Could not determine zipball_url from:\n{release_json}\n")
        log.debug(f"Fall back to {zipball_url}")

    # Make a request for the zipball
    resp = requests.get(url=zipball_url, headers=headers)

    # Filename indicated by the HTTP response
    filename = resp.headers["content-disposition"].split("filename=")[-1]
    # Location for the cached zipball
    target = platformdirs.user_cache_path("sdmx").joinpath(filename)

    # Avoid downloading if the same file is already present
    if target.exists() and not force:
        log.info(f"Use existing {target}")
        resp.close()
    else:
        # Write response content to file
        target.parent.mkdir(parents=True, exist_ok=True)
        target.write_bytes(resp.content)

    with zipfile.ZipFile(target) as zf:
        # The first name list is the top-level directory within the file
        result = target.parent.joinpath(zf.namelist()[0])

        if result.exists() and not force:
            log.info(
                f"Destination {result} exists → skip extraction.\n"
                "Remove the directory or give force=True to override"
            )
        else:
            # Unpack the entire archive
            zf.extractall(target.parent)

    return result


def _handle_validate_args(
    schema_dir: Optional[Path], version: Union[str, Version]
) -> tuple[Path, Version]:
    """Handle arguments for :func:`.install_schemas` and :func:`.validate_xml`."""
    import platformdirs

    supported = {Version["2.1"], Version["3.0.0"]}
    try:
        version = Version[version] if isinstance(version, str) else version
        assert version in supported
    except (AssertionError, KeyError):
        raise NotImplementedError(
            f"SDMX-ML version must be one of {supported}; got {version}"
        ) from None

    # If the user has no preference, download the schemas to the local cache directory
    if not schema_dir:
        schema_dir = platformdirs.user_cache_path("sdmx") / version.name
    schema_dir.mkdir(exist_ok=True, parents=True)

    return schema_dir, version


def install_schemas(
    schema_dir: Optional[Path] = None,
    version: Union[str, Version] = Version["2.1"],
) -> Path:
    """Install SDMX-ML XML Schema documents for use with :func:`.validate_xml`.

    Parameters
    ----------
    schema_dir : .Path, optional
        The directory where XSD schemas will be downloaded to. Default: a subdirectory
        named :file:`sdmx/{version}` within the :meth:`platformdirs.user_cache_path`.
    version : str or Version, optional
        The SDMX-ML schema version to install. One of :py:`Version["2.1"]` (default),
        :py:`Version["3.0.0"]`, or :class:`str` equivalent.

    Returns
    -------
    .Path
        The path containing the installed schemas. If `schema_dir` is given, the return
        value is identical to the parameter.
    """
    schema_dir, version = _handle_validate_args(schema_dir, version)

    # Copy the entire "schemas" subtree recursively
    copytree(
        _extracted_zipball(version).joinpath("schemas"), schema_dir, dirs_exist_ok=True
    )
    return schema_dir


class XMLFormat:
    """Information about an SDMX-ML format."""

    NS: Mapping[str, Optional[str]]
    _class_tag: list

    def __init__(self, model, base_ns: str, class_tag: Iterable[tuple[str, str]]):
        from sdmx import message  # noqa: F401

        self.base_ns = base_ns

        # Construct name spaces
        self.NS = {
            prefix: url if url is None else url.format(base_ns)
            for prefix, url in NS.items()
        }

        # Construct class-tag mapping
        self._class_tag = []

        # Defined in this file
        for name in CT1:
            self._class_tag.append((getattr(model, name), self.qname("str", name)))

        # Defined in this file + those passed to the constructor
        for expr, tag in chain(CT2, class_tag):
            self._class_tag.append((eval(expr), self.qname(tag)))

    @lru_cache()
    def ns_prefix(self, url) -> str:
        """Return the namespace prefix from :attr:`.NS` given its full `url`."""
        for prefix, _url in self.NS.items():
            if url == _url:
                return prefix
        raise ValueError(url)

    _NS_PATTERN = re.compile(r"(\{(?P<ns>.*)\}|(?P<ns_prefix>.*):)?(?P<localname>.*)")

    @lru_cache()
    def qname(self, ns_or_name: str, name: Optional[str] = None) -> QName:
        """Return a fully-qualified tag `name` in namespace `ns`."""
        if isinstance(ns_or_name, QName):
            # Already a QName; do nothing
            return ns_or_name

        if name is None:
            # `ns_or_name` contains the local name ("tag") and possibly a namespace
            # prefix ("ns:tag") or full namespace name ("{foo}tag")
            match = self._NS_PATTERN.fullmatch(ns_or_name)
            assert match
            name = match.group("localname")
            if prefix := match.group("ns_prefix"):
                ns = self.NS[prefix]
            elif ns := match.group("ns"):
                pass
            else:
                ns = None  # Tag without namespace
        else:
            # `ns_or_name` is the namespace prefix; `name` is the local name
            ns = self.NS[ns_or_name]

        return QName(ns, name)

    @lru_cache()
    def class_for_tag(self, tag) -> Optional[type]:
        """Return a message or model class for an XML tag."""
        qname = self.qname(tag)
        results = map(itemgetter(0), filter(lambda ct: ct[1] == qname, self._class_tag))
        try:
            return next(results)
        except StopIteration:
            return None

    @lru_cache()
    def tag_for_class(self, cls):
        """Return an XML tag for a message or model class."""
        results = map(itemgetter(1), filter(lambda ct: ct[0] is cls, self._class_tag))
        try:
            return next(results)
        except StopIteration:
            return None

    # __eq__ and __hash__ to enable lru_cache()
    def __eq__(self, other):
        return self.base_ns == other.base_ns  # pragma: no cover

    def __hash__(self):
        return hash(self.base_ns)



"""Information about the SDMX-ML 3.0 file format."""

from sdmx.model import v30

from .common import XMLFormat

FORMAT = XMLFormat(
    model=v30,
    base_ns="http://www.sdmx.org/resources/sdmxml/schemas/v3_0",
    class_tag=[
        ("model.DataflowRelationship", "str:None"),
        ("model.ObservationRelationship", "str:Observation"),
        ("model.Dataflow", "str:Dataflow"),
        ("model.MetadataAttributeDescriptor", "str:MetadataAttributeList"),
        ("model.Metadataflow", "str:Metadataflow"),
        ("model.MetadataSet", "mes:MetadataSet"),
    ]
    + [
        (f"model.{name}", f"str:{name}")
        for name in """
            ExclusiveCodeSelection
            InclusiveCodeSelection
            CodelistExtension
            DataConstraint
            GeoFeatureSetCode
            GeographicCodelist
            GeoGridCode
            GeoGridCodelist
            Hierarchy
            Measure
            MetadataAttribute
            MetadataConstraint
            ValueItem
            ValueList
        """.split()
    ],
)


def __getattr__(name):
    return getattr(FORMAT, name)



from itertools import chain
from typing import Any, Hashable, Union

import numpy as np
import pandas as pd

from sdmx import message
from sdmx.dictlike import DictLike
from sdmx.model import common, v21
from sdmx.model import v21 as model
from sdmx.model.v21 import (
    DEFAULT_LOCALE,
    AllDimensions,
    DataAttribute,
    DataSet,
    Dimension,
    DimensionComponent,
    Item,
    Observation,
    SeriesKey,
    TimeDimension,
)
from sdmx.writer.base import BaseWriter

#: Default return type for :func:`write_dataset` and similar methods. Either 'compat' or
#: 'rows'. See the ref:`HOWTO <howto-rtype>`.
DEFAULT_RTYPE = "rows"

_HAS_PANDAS_2 = pd.__version__.split(".")[0] >= "2"


writer = BaseWriter("pandas")


def to_pandas(obj, *args, **kwargs):
    """Convert an SDMX *obj* to :mod:`pandas` object(s).

    See :ref:`sdmx.writer.pandas <writer-pandas>`.
    """
    return writer.recurse(obj, *args, **kwargs)


# Functions for Python containers
@writer
def _list(obj: list, *args, **kwargs):
    """Convert a :class:`list` of SDMX objects."""
    if isinstance(obj[0], Observation):
        return write_dataset(model.DataSet(obs=obj), *args, **kwargs)
    elif isinstance(obj[0], DataSet) and len(obj) == 1:
        return writer.recurse(obj[0], *args, **kwargs)
    elif isinstance(obj[0], SeriesKey):
        assert len(args) == len(kwargs) == 0
        return write_serieskeys(obj)
    else:
        return [writer.recurse(item, *args, **kwargs) for item in obj]


@writer
def _dict(obj: dict, *args, **kwargs):
    """Convert mappings."""
    result = {k: writer.recurse(v, *args, **kwargs) for k, v in obj.items()}

    result_type = set(type(v) for v in result.values())

    if result_type <= {pd.Series, pd.DataFrame}:
        if (
            len(set(map(lambda s: s.index.name, result.values()))) == 1
            and len(result) > 1
        ):
            # Can safely concatenate these to a pd.MultiIndex'd Series.
            return pd.concat(result)
        else:
            # The individual pd.Series are indexed by different dimensions; do not
            # concatenate
            return DictLike(result)
    elif result_type == {str}:
        return pd.Series(result)
    elif result_type < {dict, DictLike}:
        return result
    elif result_type == set():
        # No results
        return pd.Series()
    else:
        raise ValueError(result_type)


@writer
def _set(obj: set, *args, **kwargs):
    """Convert :class:`set`."""
    result = {writer.recurse(o, *args, **kwargs) for o in obj}
    return result


# Functions for message classes
@writer
def write_datamessage(obj: message.DataMessage, *args, rtype=None, **kwargs):
    """Convert :class:`.DataMessage`.

    Parameters
    ----------
    rtype : 'compat' or 'rows', optional
        Data type to return; default :data:`.DEFAULT_RTYPE`. See the
        :ref:`HOWTO <howto-rtype>`.
    kwargs :
        Passed to :meth:`write_dataset` for each data set.

    Returns
    -------
    :class:`pandas.Series` or :class:`pandas.DataFrame`
        if `obj` has only one data set.
    list of (:class:`pandas.Series` or :class:`pandas.DataFrame`)
        if `obj` has more than one data set.
    """
    # Pass the message's DSD to assist datetime handling
    assert obj.dataflow
    kwargs.setdefault("dsd", obj.dataflow.structure)

    # Pass the return type and associated information
    kwargs["_rtype"] = rtype or DEFAULT_RTYPE
    if kwargs["_rtype"] == "compat":
        kwargs["_message_class"] = obj.__class__
        kwargs["_observation_dimension"] = obj.observation_dimension

    if len(obj.data) == 1:
        return writer.recurse(obj.data[0], *args, **kwargs)
    else:
        return [writer.recurse(ds, *args, **kwargs) for ds in obj.data]


@writer
def write_structuremessage(obj: message.StructureMessage, include=None, **kwargs):
    """Convert :class:`.StructureMessage`.

    Parameters
    ----------
    obj : .StructureMessage
    include : iterable of str or str, optional
        One or more of the attributes of the StructureMessage ('category_scheme',
        'codelist', etc.) to transform.
    kwargs :
        Passed to :meth:`write` for each attribute.

    Returns
    -------
    .DictLike
        Keys are StructureMessage attributes; values are pandas objects.
    """
    all_contents = {
        "category_scheme",
        "codelist",
        "concept_scheme",
        "constraint",
        "dataflow",
        "structure",
        "organisation_scheme",
    }

    # Handle arguments
    if include is None:
        attr_set = all_contents
    else:
        attr_set = set([include] if isinstance(include, str) else include)
        # Silently discard invalid names
        attr_set &= all_contents
    attrs = sorted(attr_set)

    result: DictLike[str, Union[pd.Series, pd.DataFrame]] = DictLike()
    for a in attrs:
        dl = writer.recurse(getattr(obj, a), **kwargs)
        if len(dl):
            # Only add non-empty elements
            result[a] = dl

    return result


# Functions for model classes


@writer
def _c(obj: model.Component):
    """Convert :class:`.Component`."""
    # Raises AttributeError if the concept_identity is missing
    return str(obj.concept_identity.id)  # type: ignore


@writer
def _cc(obj: model.ContentConstraint, **kwargs):
    """Convert :class:`.ContentConstraint`."""
    return {
        i: writer.recurse(cr, **kwargs) for i, cr in enumerate(obj.data_content_region)
    }


@writer
def _cr(obj: model.CubeRegion, **kwargs):
    """Convert :class:`.CubeRegion`."""
    result: DictLike[str, pd.Series] = DictLike()
    for dim, ms in obj.member.items():
        result[dim.id] = pd.Series(
            [writer.recurse(sv, **kwargs) for sv in ms.values], name=dim.id
        )
    return result


@writer
def _rp(obj: model.RangePeriod, **kwargs):
    """Convert :class:`.RangePeriod`."""
    return f"{obj.start.period}–{obj.end.period}"


@writer
def write_dataset(  # noqa: C901 TODO reduce complexity 12 → ≤10
    obj: common.BaseDataSet,
    attributes="",
    dtype=np.float64,
    constraint=None,
    datetime=False,
    **kwargs,
):
    """Convert :class:`~.DataSet`.

    See the :ref:`walkthrough <datetime>` for examples of using the `datetime` argument.

    Parameters
    ----------
    obj : :class:`~.DataSet` or iterable of :class:`Observation <.BaseObservation>`
    attributes : str
        Types of attributes to return with the data. A string containing zero or more
        of:

        - ``'o'``: attributes attached to each :class:`Observation <.BaseObservation>` .
        - ``'s'``: attributes attached to any (0 or 1) :class:`~.SeriesKey` associated
          with each Observation.
        - ``'g'``: attributes attached to any (0 or more) :class:`~.GroupKey` associated
          with each Observation.
        - ``'d'``: attributes attached to the :class:`~.DataSet` containing the
          Observations.

    dtype : str or :class:`numpy.dtype` or None
        Datatype for values. If None, do not return the values of a series. In this
        case, `attributes` must not be an empty string so that some attribute is
        returned.
    constraint : .ContentConstraint, optional
        If given, only Observations included by the *constraint* are returned.
    datetime : bool or str or or .Dimension or dict, optional
        If given, return a DataFrame with a :class:`~pandas.DatetimeIndex` or
        :class:`~pandas.PeriodIndex` as the index and all other dimensions as columns.
        Valid `datetime` values include:

        - :class:`bool`: if :obj:`True`, determine the time dimension automatically by
          detecting a :class:`~.TimeDimension`.
        - :class:`str`: ID of the time dimension.
        - :class:`~.Dimension`: the matching Dimension is the time dimension.
        - :class:`dict`: advanced behaviour. Keys may include:

          - **dim** (:class:`~.Dimension` or :class:`str`): the time dimension or its
            ID.
          - **axis** (`{0 or 'index', 1 or 'columns'}`): axis on which to place the time
            dimension (default: 0).
          - **freq** (:obj:`True` or :class:`str` or :class:`~.Dimension`): produce
            :class:`pandas.PeriodIndex`. If :class:`str`, the ID of a Dimension
            containing a frequency specification. If a Dimension, the specified
            dimension is used for the frequency specification.

            Any Dimension used for the frequency specification is does not appear in the
            returned DataFrame.

    Returns
    -------
    :class:`pandas.DataFrame`
        - if `attributes` is not ``''``, a data frame with one row per Observation,
          ``value`` as the first column, and additional columns for each attribute;
        - if `datetime` is given, various layouts as described above; or
        - if `_rtype` (passed from :func:`write_datamessage`) is 'compat', various
          layouts as described in the :ref:`HOWTO <howto-rtype>`.
    :class:`pandas.Series` with :class:`pandas.MultiIndex`
        Otherwise.
    """
    # If called directly on a DataSet (rather than a parent DataMessage), cannot
    # determine the "dimension at observation level"
    rtype = kwargs.setdefault("_rtype", "rows")

    # Validate attributes argument
    attributes = attributes or ""
    try:
        attributes = attributes.lower()
    except AttributeError:
        raise TypeError("'attributes' argument must be str")

    if rtype == "compat" and kwargs["_observation_dimension"] is not AllDimensions:
        # Cannot return attributes in this case
        attributes = ""
    elif set(attributes) - {"o", "s", "g", "d"}:
        raise ValueError(f"attributes must be in 'osgd'; got {attributes}")

    # Iterate on observations
    data: dict[Hashable, dict[str, Any]] = {}
    for observation in obj.obs:
        # Check that the Observation is within the constraint, if any
        key = observation.key.order()
        if constraint and key not in constraint:
            continue

        # Add value and attributes
        row = {}
        if dtype:
            row["value"] = observation.value
        if attributes:
            # Add the combined attributes from observation, series- and group keys
            row.update(observation.attrib)
        if "d" in attributes and isinstance(obj, v21.DataSet):
            # Add the attributes of the data set
            row.update(obj.attrib)

        data[tuple(map(str, key.get_values()))] = row

    # NB mypy errors here on CI, but not locally
    result: Union[pd.Series, pd.DataFrame] = pd.DataFrame.from_dict(
        data,
        orient="index",  # type: ignore [arg-type]
    )

    if len(result):
        result.index.names = observation.key.order().values.keys()
        if dtype:
            try:
                result["value"] = result["value"].astype(dtype)
            except ValueError:
                # Attempt to handle locales in which LC_NUMERIC.decimal_point is ","
                # TODO Make this more robust by inferring and changing locale settings
                result["value"] = result["value"].str.replace(",", ".").astype(dtype)
            if not attributes:
                result = result["value"]

    # Reshape for compatibility with v0.9
    result, datetime, kwargs = _dataset_compat(result, datetime, kwargs)
    # Handle the datetime argument, if any
    return _maybe_convert_datetime(result, datetime, obj=obj, **kwargs)


def _dataset_compat(df, datetime, kwargs):
    """Helper for :meth:`.write_dataset` 0.9 compatibility."""
    rtype = kwargs.pop("_rtype")
    if rtype != "compat":
        return df, datetime, kwargs  # Do nothing

    # Remove compatibility arguments from kwargs
    kwargs.pop("_message_class")
    obs_dim = kwargs.pop("_observation_dimension")

    if isinstance(obs_dim, list) and len(obs_dim) == 1:
        # Unwrap a length-1 list
        obs_dim = obs_dim[0]

    if obs_dim in (AllDimensions, None):
        pass  # Do nothing
    elif isinstance(obs_dim, TimeDimension):
        # Don't modify *df*; only change arguments so that _maybe_convert_datetime
        # performs the desired changes
        if datetime is False or datetime is True:
            # Either datetime is not given, or True without specifying a dimension;
            # overwrite
            datetime = obs_dim
        elif isinstance(datetime, dict):
            # Dict argument; ensure the 'dim' key is the same as obs_dim
            if datetime.setdefault("dim", obs_dim) != obs_dim:
                msg = (
                    f"datetime={datetime} conflicts with rtype='compat' and"
                    f" {obs_dim} at observation level"
                )
                raise ValueError(msg)
        else:
            assert datetime == obs_dim, (datetime, obs_dim)
    elif isinstance(obs_dim, DimensionComponent):
        # Pivot all levels except the observation dimension
        df = df.unstack([n for n in df.index.names if n != obs_dim.id])
    else:
        # E.g. some JSON messages have two dimensions at the observation level;
        # behaviour is unspecified here, so do nothing.
        pass

    return df, datetime, kwargs


def _maybe_convert_datetime(df, arg, obj, dsd=None):  # noqa: C901  TODO reduce complexity 23 → ≤10
    """Helper for :meth:`.write_dataset` to handle datetime indices.

    Parameters
    ----------
    df : pandas.DataFrame
    arg : dict
        From the `datetime` argument to :meth:`write_dataset`.
    obj :
        From the `obj` argument to :meth:`write_dataset`.
    dsd: ~.DataStructureDefinition, optional
    """
    # TODO Simplify this method to reduce its McCabe complexity from 23 to <= 13

    # Check argument values
    param = dict(dim=None, axis=0, freq=False)

    if not arg:
        return df  # False, None, empty dict → no datetime conversion
    elif isinstance(arg, str):
        param["dim"] = arg
    elif isinstance(arg, DimensionComponent):
        param["dim"] = arg.id
    elif isinstance(arg, dict):
        extra_keys = set(arg.keys()) - set(param.keys())
        if extra_keys:
            raise ValueError(extra_keys)
        param.update(arg)
    elif isinstance(arg, bool):
        pass  # True
    else:
        raise ValueError(arg)

    def _get(kind: str):
        """Return an appropriate list of dimensions or attributes."""
        if len(getattr(obj.structured_by, kind).components):
            return getattr(obj.structured_by, kind).components
        elif dsd:
            return getattr(dsd, kind).components
        else:
            return []

    # Determine time dimension
    if not param["dim"]:
        for dim in filter(lambda d: isinstance(d, TimeDimension), _get("dimensions")):
            param["dim"] = dim
            break
    if not param["dim"]:
        raise ValueError(f"no TimeDimension in {_get('dimensions')}")

    # Unstack all but the time dimension and convert
    other_dims = list(filter(lambda d: d != param["dim"], df.index.names))
    df = df.unstack(other_dims)
    # Only provide format in pandas >= 2.0.0
    kw = dict(format="mixed") if _HAS_PANDAS_2 else {}
    df.index = pd.to_datetime(df.index, **kw)

    # Convert to a PeriodIndex with a particular frequency
    if freq := param["freq"]:
        try:
            # A frequency string recognized by pandas.PeriodDtype
            if isinstance(freq, str):
                freq = pd.PeriodDtype(freq=freq).freq
        except ValueError:
            # ID of a Dimension; Attribute; or column of `df`
            result = None
            for component in chain(
                _get("dimensions"), _get("attributes"), map(Dimension, df.columns.names)
            ):
                if component.id == freq:
                    freq = result = component
                    break

            if not result:
                raise ValueError(freq)

        if isinstance(freq, Dimension):
            # Retrieve Dimension values from pd.MultiIndex level
            level = freq.id
            i = df.columns.names.index(level)
            values = set(df.columns.levels[i])

            if len(values) > 1:
                raise ValueError(
                    f"cannot convert to PeriodIndex with non-unique freq={sorted(values)}"
                )

            # Store the unique value
            freq = values.pop()

            # Remove the index level
            df.columns = df.columns.droplevel(i)
        elif isinstance(freq, DataAttribute):  # pragma: no cover
            raise NotImplementedError

        df.index = df.index.to_period(freq=freq)

    if param["axis"] in {1, "columns"}:
        # Change axis
        df = df.transpose()

    return df


@writer
def _dd(obj: model.DimensionDescriptor):
    """Convert :class:`.DimensionDescriptor`."""
    return writer.recurse(obj.components)


@writer
def write_itemscheme(obj: model.ItemScheme, locale=DEFAULT_LOCALE):
    """Convert :class:`.ItemScheme`.

    Parameters
    ----------
    locale : str, optional
        Locale for names to return.

    Returns
    -------
    pandas.Series or pandas.DataFrame
    """
    items = {}
    seen: set[Item] = set()

    def add_item(item):
        """Recursive helper for adding items."""
        # Track seen items
        if item in seen:
            return
        else:
            seen.add(item)

        items[item.id] = dict(
            # Localized name
            name=item.name.localized_default(locale),
            # Parent ID
            parent=item.parent.id if isinstance(item.parent, item.__class__) else "",
        )

        # Add this item's children, recursively
        for child in item.child:
            add_item(child)

    for item in obj:
        add_item(item)

    # Convert to DataFrame
    result: Union[pd.DataFrame, pd.Series] = pd.DataFrame.from_dict(
        items,
        orient="index",
        dtype=object,  # type: ignore [arg-type]
    ).rename_axis(obj.id, axis="index")

    if len(result) and not result["parent"].str.len().any():
        # 'parent' column is empty; convert to pd.Series and rename
        result = result["name"].rename(obj.name.localized_default(locale))

    return result


@writer
def _mv(obj: model.MemberValue):
    return obj.value


@writer
def _mds(obj: model.MetadataSet, **kwargs):
    raise NotImplementedError(f"write {type(obj).__name__} to pandas")


@writer
def _na(obj: model.NameableArtefact, **kwargs):
    """Fallback for NameableArtefact: only its name."""
    return str(obj.name)


def write_serieskeys(obj):
    result = []
    for sk in obj:
        result.append({dim: kv.value for dim, kv in sk.order().values.items()})
    # TODO perhaps return as a pd.MultiIndex if that is more useful
    return pd.DataFrame(result)



"""SDMX-ML v2.1 writer."""

# Contents of this file are organized in the order:
#
# - Utility methods and global variables.
# - writer functions for sdmx.message classes, in the same order as message.py
# - writer functions for sdmx.model classes, in the same order as model.py
import logging
from datetime import datetime
from typing import Iterable, Literal, MutableMapping, Optional

from lxml import etree
from lxml.builder import ElementMaker

import sdmx.urn
from sdmx import message
from sdmx.format.xml.v21 import NS, qname, tag_for_class
from sdmx.model import common, v21
from sdmx.model import v21 as model
from sdmx.util import ucfirst
from sdmx.writer.base import BaseWriter

_element_maker = ElementMaker(nsmap={k: v for k, v in NS.items() if v is not None})

log = logging.getLogger(__name__)

writer = BaseWriter("XML")


def Element(name, *args, **kwargs):
    # Remove None
    kwargs = dict(filter(lambda kv: kv[1] is not None, kwargs.items()))

    return _element_maker(qname(name), *args, **kwargs)


def to_xml(obj, **kwargs):
    """Convert an SDMX *obj* to SDMX-ML.

    Parameters
    ----------
    kwargs
        Passed to :meth:`lxml.etree.to_string`, e.g. `pretty_print` = :obj:`True`.

    Raises
    ------
    NotImplementedError
        If writing specific objects to SDMX-ML has not been implemented in :mod:`sdmx`.
    """
    kwargs.setdefault("encoding", "utf-8")
    kwargs.setdefault("xml_declaration", True)
    return etree.tostring(writer.recurse(obj), **kwargs)


RefStyle = Literal["Ref", "URN"]


def reference(obj, parent=None, tag=None, *, style: RefStyle):
    """Write a reference to `obj`.

    .. todo:: Currently other functions in :mod:`.writer.xml` all pass the `style`
       argument to this function. As an enhancement, allow user or automatic selection
       of different reference styles.
    """
    tag = tag or tag_for_class(obj.__class__)

    elem = Element(tag)

    # assert style
    if style == "URN":
        ref = Element(":URN", obj.urn)
    elif style == "Ref":
        # Element attributes
        attrib = dict(id=obj.id)

        # Identify a maintainable artifact; either `obj` or its `parent`
        if isinstance(obj, model.MaintainableArtefact):
            ma = obj
            attrib.update(version=obj.version)
        else:
            try:
                # Get the ItemScheme for an Item
                parent = parent or obj.get_scheme()
            except AttributeError:  # pragma: no cover
                # No `parent` and `obj` is not an Item with a .get_scheme() method
                # NB this does not occur in the test suite
                pass

            if not parent:
                raise NotImplementedError(
                    f"Cannot write reference to {obj!r} without parent"
                )

            ma = parent
            attrib.update(
                maintainableParentVersion=ma.version,
                maintainableParentID=ma.id,
            )

        attrib.update(
            agencyID=getattr(ma.maintainer, "id", None),
            package=model.PACKAGE[type(ma).__name__],
        )

        # "class" attribute: either the type of `obj`, or the item type of an ItemScheme
        for candidate in (type(obj), getattr(type(ma), "_Item", None)):
            try:
                attrib["class"] = etree.QName(tag_for_class(candidate)).localname
                break
            except ValueError:
                pass

        ref = Element(":Ref", **attrib)

    elem.append(ref)
    return elem


# Writers for sdmx.message classes


@writer
def _dm(obj: message.DataMessage):
    """DataMessage, including MetadataMessage."""
    # Identify root tag
    if len(obj.data) and isinstance(
        obj.data[0],
        (model.StructureSpecificDataSet, model.StructureSpecificTimeSeriesDataSet),
    ):
        tag = "mes:StructureSpecificData"
    else:
        tag = tag_for_class(type(obj))

    # Create the root element
    elem = Element(tag)

    header = writer.recurse(obj.header)
    elem.append(header)

    # Set of DSDs already referenced in the header
    structures = set()

    for ds in obj.data:
        attrib = dict()
        dsd_ref = None

        # Add any new DSD reference to header
        if ds.structured_by and id(ds.structured_by) not in structures:
            attrib["structureID"] = ds.structured_by.id

            # Reference by URN if possible, otherwise with a <Ref> tag
            style: RefStyle = "URN" if ds.structured_by.urn else "Ref"
            dsd_ref = reference(ds.structured_by, tag="com:Structure", style=style)

            if isinstance(obj.observation_dimension, model.DimensionComponent):
                attrib["dimensionAtObservation"] = obj.observation_dimension.id

            header.append(Element("mes:Structure", **attrib))
            header[-1].append(dsd_ref)

            # Record this object so it is not added a second time
            structures.add(id(ds.structured_by))

        # Add data
        elem.append(writer.recurse(ds))

    if obj.footer:
        elem.append(writer.recurse(obj.footer))

    return elem


@writer
def _sm(obj: message.StructureMessage):
    # Store a reference to the overal Message for writing references
    setattr(writer, "_message", obj)

    elem = Element("mes:Structure")

    # Empty header element
    elem.append(writer.recurse(obj.header))

    structures = Element("mes:Structures")
    elem.append(structures)

    for attr, tag in [
        # Order is important here to avoid forward references
        ("organisation_scheme", "OrganisationSchemes"),
        ("dataflow", "Dataflows"),
        ("category_scheme", "CategorySchemes"),
        ("categorisation", "Categorisations"),
        ("codelist", "Codelists"),
        ("concept_scheme", "Concepts"),
        ("structure", "DataStructures"),
        ("constraint", "Constraints"),
        ("metadatastructure", "MetadataStructures"),
        ("provisionagreement", "ProvisionAgreements"),
    ]:
        coll = getattr(obj, attr)
        if not len(coll):
            continue
        container = Element(f"str:{tag}")

        for s in filter(lambda s: not s.is_external_reference, coll.values()):
            container.append(writer.recurse(s))

        if len(container):
            structures.append(container)

    if obj.footer:
        elem.append(writer.recurse(obj.footer))

    return elem


@writer
def _em(obj: message.ErrorMessage):
    elem = Element("mes:Error")
    elem.append(writer.recurse(obj.header))

    if obj.footer:
        elem.append(writer.recurse(obj.footer))

    return elem


@writer
def _header(obj: message.Header):
    elem = Element(
        "mes:Header",
        # Mandatory child elements of mes:Header
        Element("mes:ID", obj.id or "none"),
        Element("mes:Test", str(obj.test).lower()),
        Element("mes:Prepared", (obj.prepared or datetime.now()).isoformat()),
        writer.recurse(obj.sender or common.Agency(id="none"), _tag="mes:Sender"),
    )
    # Optional child elements
    if obj.receiver:
        elem.append(writer.recurse(obj.receiver, _tag="mes:Receiver"))
    if obj.source:
        elem.extend(i11lstring(obj.source, "mes:Source"))
    return elem


@writer
def _footer(obj: message.Footer):
    elem = Element("footer:Footer")

    attrs = dict()
    if obj.code:
        attrs["code"] = str(obj.code)
    if obj.severity:
        attrs["severity"] = str(obj.severity)

    mes = Element("footer:Message", **attrs)
    elem.append(mes)

    for text in obj.text:
        mes.extend(i11lstring(text, "com:Text"))

    return elem


# Writers for sdmx.model classes
# §3.2: Base structures


def i11lstring(obj, name) -> list[etree._Element]:
    """InternationalString.

    Returns a list of elements with name `name`.
    """
    elems = []

    for locale, label in obj.localizations.items():
        child = Element(name, label)
        child.set(qname("xml", "lang"), locale)
        elems.append(child)

    return elems


@writer
def _a(obj: model.Annotation):
    elem = Element("com:Annotation")
    if obj.id:
        elem.attrib["id"] = obj.id
    if obj.title:
        elem.append(Element("com:AnnotationTitle", obj.title))
    if obj.type:
        elem.append(Element("com:AnnotationType", obj.type))
    elem.extend(i11lstring(obj.text, "com:AnnotationText"))
    return elem


def annotable(obj: common.AnnotableArtefact, *args, **kwargs) -> etree._Element:
    # Determine tag
    tag = kwargs.pop("_tag", tag_for_class(obj.__class__))
    if tag is None:  # pragma: no cover
        raise NotImplementedError(f"Write {obj.__class__} to SDMX-ML")

    # Write Annotations
    e_anno = Element("com:Annotations", *[writer.recurse(a) for a in obj.annotations])
    if len(e_anno):
        args = (e_anno,) + args

    try:
        return Element(tag, *args, **kwargs)
    except AttributeError:  # pragma: no cover
        print(repr(obj), tag, kwargs)
        raise


def identifiable(obj: common.IdentifiableArtefact, *args, **kwargs) -> etree._Element:
    """Write :class:`.IdentifiableArtefact`.

    Unless the keyword argument `_with_urn` is :data:`False`, a URN is generated for
    objects lacking one, and forwarded to :func:`annotable`
    """
    kwargs.setdefault("id", obj.id)
    try:
        with_urn = kwargs.pop("_with_urn", True)
        urn = obj.urn or (
            sdmx.urn.make(obj, kwargs.pop("parent", None)) if with_urn else None
        )
        if urn:
            kwargs.setdefault("urn", urn)
    except (AttributeError, ValueError):
        pass
    return annotable(obj, *args, **kwargs)


def nameable(obj: common.NameableArtefact, *args, **kwargs) -> etree._Element:
    return identifiable(
        obj,
        *i11lstring(obj.name, "com:Name"),
        *i11lstring(obj.description, "com:Description"),
        *args,
        **kwargs,
    )


def maintainable(obj: common.MaintainableArtefact, *args, **kwargs) -> etree._Element:
    # MaintainableArtefact attributes
    kwargs.setdefault("isExternalReference", str(obj.is_external_reference).lower())
    kwargs.setdefault("isFinal", str(obj.is_final).lower())
    kwargs.setdefault("agencyID", getattr(obj.maintainer, "id", None))

    # VersionableArtefact attributes (no separate method)
    kwargs.setdefault("version", str(obj.version))
    # TODO Also write valid_from, valid_to

    return nameable(obj, *args, **kwargs)


# §3.5: Item Scheme


@writer
def _item(obj: model.Item, **kwargs):
    elem = nameable(obj, **kwargs)

    # Reference to parent Item
    if isinstance(obj.parent, obj.__class__):
        elem.append(Element("str:Parent", Element(":Ref", id=obj.parent.id)))

    if isinstance(obj, common.Organisation):
        elem.extend(writer.recurse(c) for c in obj.contact)

    return elem


@writer
def _is(obj: model.ItemScheme):
    kw = dict()
    if obj.is_partial is not None:
        kw["isPartial"] = str(obj.is_partial).lower()
    elem = maintainable(obj, **kw)

    # Pass _with_urn to identifiable(): don't generate URNs for Items in `obj` which do
    # not already have them
    elem.extend(writer.recurse(i, _with_urn=False) for i in obj.items.values())
    return elem


# §3.6: Structure


@writer
def _facet(obj: model.Facet):
    attrib: MutableMapping[str, str] = dict()
    try:
        attrib.update(textType=ucfirst(obj.value_type.name))
    except AttributeError:  # pragma: no cover
        pass
    return Element("str:TextFormat", **attrib)


@writer
def _rep(obj: common.Representation, tag, style="URN"):
    elem = Element(f"str:{tag}")
    if obj.enumerated is not None:
        elem.append(reference(obj.enumerated, tag="str:Enumeration", style=style))
    if obj.non_enumerated:
        elem.extend(writer.recurse(facet) for facet in obj.non_enumerated)
    return elem


# §4.4: Concept Scheme


@writer
def _concept(obj: model.Concept, **kwargs):
    elem = _item(obj, **kwargs)

    if obj.core_representation:
        elem.append(writer.recurse(obj.core_representation, "CoreRepresentation"))

    return elem


# §4.6: Organisations


@writer
def _contact(obj: model.Contact):
    return Element(
        "str:Contact",
        *i11lstring(obj.name, "com:Name"),
        *i11lstring(obj.org_unit, "str:Department"),
        *i11lstring(obj.responsibility, "str:Role"),
        *([Element("str:Telephone", obj.telephone)] if obj.telephone else []),
        *[Element("str:URI", value) for value in obj.uri],
        *[Element("str:Email", value) for value in obj.email],
    )


# §3.3: Basic Inheritance


@writer
def _component(obj: model.Component, dsd, *, attrib: Optional[dict] = None):
    child = []
    attrib = attrib or dict()

    try:
        child.append(
            reference(obj.concept_identity, tag="str:ConceptIdentity", style="Ref")
        )
    except AttributeError:  # pragma: no cover
        pass  # concept_identity is None

    try:
        child.append(
            writer.recurse(obj.local_representation, "LocalRepresentation", style="Ref")
        )
    except NotImplementedError:
        pass  # None

    if isinstance(obj, model.DataAttribute) and obj.usage_status:
        child.append(writer.recurse(obj.related_to, dsd))

        # assignmentStatus attribute
        if obj.usage_status:
            attrib["assignmentStatus"] = obj.usage_status.name.title()
    elif isinstance(obj, model.Dimension):
        # position attribute
        attrib["position"] = str(obj.order)

    return identifiable(obj, *child, **attrib)


@writer
def _cl(obj: model.ComponentList, *args):
    return identifiable(obj, *[writer.recurse(c, *args) for c in obj.components])


# §4.5: CategoryScheme


@writer
def _cat(obj: model.Categorisation):
    elem = maintainable(obj)
    # If .reader.xml.v21._ref did not resolve a ref while reading (e.g. ref to a non-
    # standard class like PublicationTable), this may be a dict → don't write. Generates
    # XSD-invalid SDMX-ML; but occurs only when the input SDMX-ML was invalid anyway.
    if isinstance(obj.artefact, common.IdentifiableArtefact):
        elem.append(reference(obj.artefact, tag="str:Source", style="Ref"))
    elem.append(reference(obj.category, tag="str:Target", style="Ref"))
    return elem


# §10.3: Constraints


@writer
def _dk(obj: model.DataKey):
    elem = Element("str:Key", isIncluded=str(obj.included).lower())
    for value_for, cv in obj.key_value.items():
        elem.append(Element("com:KeyValue", id=value_for.id))
        elem[-1].append(Element("com:Value", cv.value))
    return elem


@writer
def _dks(obj: model.DataKeySet):
    return Element(
        "str:DataKeySet",
        *[writer.recurse(dk) for dk in obj.keys],
        isIncluded=str(obj.included).lower(),
    )


@writer
def _mv(obj: model.MemberValue):
    return Element("com:Value", obj.value)


@writer
def _rp(obj: model.RangePeriod):
    return Element("com:TimeRange", writer.recurse(obj.start), writer.recurse(obj.end))


@writer
def _period(obj: common.Period):
    """Includes :class:`.v21.StartPeriod` and :class:`.v21.EndPeriod`."""
    return Element(
        f"com:{obj.__class__.__name__}",
        # `period` attribute as text
        obj.period.isoformat(),
        isInclusive=str(obj.is_inclusive).lower(),
    )


@writer
def _ms(obj: model.MemberSelection):
    tag = {
        model.Dimension: "KeyValue",
        model.DataAttribute: "Attribute",
    }[type(obj.values_for)]

    return Element(
        f"com:{tag}", *[writer.recurse(v) for v in obj.values], id=obj.values_for.id
    )


@writer
def _cr(obj: model.CubeRegion):
    return Element(
        "str:CubeRegion",
        *[writer.recurse(ms) for ms in obj.member.values()],
        include=str(obj.included).lower(),
    )


@writer
def _cc(obj: model.ContentConstraint):
    assert obj.role is not None
    elem = maintainable(
        obj, type=obj.role.role.name.replace("allowable", "allowed").title()
    )

    # Constraint attachment: written before data_content_keys or data_content_region
    for ca in obj.content:
        elem.append(Element("str:ConstraintAttachment"))
        elem[-1].append(reference(ca, style="Ref"))

    # NB this is a property of Constraint, not ContentConstraint, so the code should be
    #    copied/reused for AttachmentConstraint.
    if obj.data_content_keys is not None:
        elem.append(writer.recurse(obj.data_content_keys))

    elem.extend(writer.recurse(dcr) for dcr in obj.data_content_region)

    return elem


# §5.2: Data Structure Definition


@writer
def _nsr(obj: model.NoSpecifiedRelationship, *args):
    elem = Element("str:AttributeRelationship")
    elem.append(Element("str:None"))
    return elem


@writer
def _pmr(obj: model.PrimaryMeasureRelationship, dsd: model.DataStructureDefinition):
    elem = Element("str:AttributeRelationship")
    elem.append(Element("str:PrimaryMeasure"))
    try:
        ref_id = dsd.measures[0].id
    except IndexError:
        ref_id = "(not implemented)"  # MeasureDescriptor is empty
    elem[-1].append(Element(":Ref", id=ref_id))
    return elem


@writer
def _dr(obj: common.DimensionRelationship, *args):
    elem = Element("str:AttributeRelationship")
    for dim in obj.dimensions:
        elem.append(Element("str:Dimension"))
        elem[-1].append(Element(":Ref", id=dim.id))
    return elem


@writer
def _gr(obj: common.GroupRelationship, *args):
    elem = Element("str:AttributeRelationship")
    elem.append(Element("str:Group"))
    elem[-1].append(Element(":Ref", id=getattr(obj.group_key, "id", None)))
    return elem


@writer
def _gdd(obj: model.GroupDimensionDescriptor):
    elem = identifiable(obj)
    for dim in obj.components:
        elem.append(Element("str:GroupDimension"))
        elem[-1].append(Element("str:DimensionReference"))
        elem[-1][0].append(Element(":Ref", id=dim.id))
    return elem


@writer
def _dsd(obj: model.DataStructureDefinition):
    elem = maintainable(obj)
    elem.append(Element("str:DataStructureComponents"))

    # Write in a specific order
    elem[-1].append(writer.recurse(obj.dimensions, None))
    for group in obj.group_dimensions.values():
        elem[-1].append(writer.recurse(group))
    elem[-1].append(writer.recurse(obj.attributes, obj))
    elem[-1].append(writer.recurse(obj.measures, None))

    return elem


@writer
def _dfd(obj: model.DataflowDefinition):
    elem = maintainable(obj)
    elem.append(reference(obj.structure, tag="str:Structure", style="Ref"))
    return elem


# §5.4: Data Set


def _av(name: str, obj: Iterable[model.AttributeValue]):
    elements = []
    for av in obj:
        assert av.value_for
        elements.append(Element("gen:Value", id=av.value_for.id, value=av.value))
    return Element(name, *elements)


def _kv(name: str, obj: Iterable[model.KeyValue]):
    elements = []
    for kv in obj:
        assert kv.value_for
        elements.append(Element("gen:Value", id=kv.value_for.id, value=str(kv.value)))
    return Element(name, *elements)


@writer
def _sk(obj: model.SeriesKey):
    elem = []

    elem.append(_kv("gen:SeriesKey", obj))
    if len(obj.attrib):
        elem.append(_av("gen:Attributes", obj.attrib.values()))

    return tuple(elem)


@writer
def _obs(obj: model.Observation, struct_spec=False):
    if struct_spec:
        obs_attrs = {}
        for key, av in obj.attached_attribute.items():
            obs_attrs[key] = str(av.value)
        if obj.value is not None:
            if obj.value_for is None:
                raise ValueError(
                    "Observation.value_for is None when writing structure-specific data"
                )
            # NB this is usually OBS_VALUE, but not necessarily; see #67.
            value_key = obj.value_for.id
            obs_attrs[value_key] = str(obj.value)
        if obj.dimension:
            for key, dv in obj.dimension.values.items():
                obs_attrs[key] = str(dv.value)

        return Element(":Obs", **obs_attrs)

    elem = Element("gen:Obs")

    if obj.dimension:
        if len(obj.dimension) == 1:
            # Observation in a series; at most one dimension given by the Key
            elem.append(
                Element("gen:ObsDimension", value=obj.dimension.values[0].value)
            )
        else:
            # Top-level observation, not associated with a SeriesKey
            elem.append(_kv("gen:ObsKey", obj.dimension))

    elem.append(Element("gen:ObsValue", value=str(obj.value)))

    if len(obj.attached_attribute):
        elem.append(_av("gen:Attributes", obj.attached_attribute.values()))

    return elem


@writer
def _ds(obj: model.DataSet):
    if len(obj.group):
        raise NotImplementedError("to_xml() for DataSet with groups")

    attrib = dict()
    if obj.action:
        attrib["action"] = str(obj.action)
    if obj.structured_by:
        attrib["structureRef"] = obj.structured_by.id
    elem = Element("mes:DataSet", **attrib)

    # AttributeValues attached to the data set
    if len(obj.attrib):
        elem.append(_av("gen:Attributes", obj.attrib.values()))

    obs_to_write = set(map(id, obj.obs))

    struct_spec = isinstance(
        obj, (model.StructureSpecificDataSet, model.StructureSpecificTimeSeriesDataSet)
    )

    for sk, observations in obj.series.items():
        if struct_spec:
            series_attrs = {}
            for key, sk_dim in sk.values.items():
                series_attrs[key] = str(sk_dim.value)
            for key, sk_att in sk.attrib.items():
                series_attrs[key] = str(sk_att.value)
            elem.append(Element(":Series", **series_attrs))
        else:
            elem.append(Element("gen:Series"))
            elem[-1].extend(writer.recurse(sk))
        elem[-1].extend(
            writer.recurse(obs, struct_spec=struct_spec) for obs in observations
        )
        obs_to_write -= set(map(id, observations))

    # Observations not in any series
    for obs in filter(lambda o: id(o) in obs_to_write, obj.obs):
        elem.append(writer.recurse(obs, struct_spec=struct_spec))

    return elem


# SDMX 2.1 §7.3: Metadata Structure Definition


@writer
def _mdsd(obj: v21.MetadataStructureDefinition):
    msc = Element(
        "str:MetadataStructureComponents",
        *[writer.recurse(mdt, obj) for mdt in obj.target.values()],
        *[writer.recurse(rs) for rs in obj.report_structure.values()],
    )

    return maintainable(obj, msc)


@writer
def _rs(obj: v21.ReportStructure):
    elem = _cl(obj, None)
    elem.extend(
        [
            Element("str:MetadataTarget", Element(":Ref", id=mdt.id))
            for mdt in obj.report_for
        ]
    )
    return elem


@writer
def _mda(obj: v21.MetadataAttribute, *args):
    # MetadataAttribute class properties
    attrib = dict()
    if obj.is_presentational is not None:
        attrib["isPresentational"] = str(obj.is_presentational).lower()
    if obj.min_occurs is not None:
        attrib["minOccurs"] = str(obj.min_occurs)
    if obj.max_occurs is not None:
        attrib["maxOccurs"] = str(obj.max_occurs)

    # Use the generic _component function to handle several common features
    elem = _component(obj, *args, attrib=attrib)

    # Recurse children
    elem.extend([writer.recurse(mda, *args) for mda in obj.child])

    return elem


@writer
def _iot(obj: v21.IdentifiableObjectTarget, *args):
    # IdentifiableObjectTarget class properties
    attrib: MutableMapping[str, str] = dict()
    if obj.object_type:
        attrib.update(objectType=str(obj.object_type.__name__))

    # Use the generic _component function to handle several common features
    return _component(obj, *args, attrib=attrib)


@writer
def _rpt(obj: v21.ReportPeriodTarget, *args):
    elem = _component(obj, *args)
    # Do not write "id" attribute
    elem.attrib.pop("id")
    return elem


# SDMX 2.1 §7.4: Metadata Set


@writer
def _mds(obj: model.MetadataSet):
    attrib = {}
    if obj.structured_by:
        attrib["structureRef"] = obj.structured_by.id
    return Element(
        "mes:MetadataSet", *[writer.recurse(mdr) for mdr in obj.report], **attrib
    )


@writer
def _mdr(obj: model.MetadataReport):
    # TODO Write the id=… attribute
    elem = Element("md:Report")

    if obj.target:  # pragma: no cover
        elem.append(writer.recurse(obj.target))
    if obj.attaches_to:
        elem.append(writer.recurse(obj.attaches_to))

    elem.append(
        Element("md:AttributeSet", *[writer.recurse(ra) for ra in obj.metadata])
    )

    return elem


@writer
def _tok(obj: model.TargetObjectKey):
    # TODO Write the id=… attribute
    return Element(
        "md:Target", *[writer.recurse(tov) for tov in obj.key_values.values()]
    )


@writer
def _tov(obj: model.TargetObjectValue):
    if isinstance(obj.value_for, str):
        # NB value_for should be MetadataAttribute, but currently not due to limitations
        #    of .reader.xml.v21
        id_: str = obj.value_for
    else:  # pragma: no cover
        id_ = obj.value_for.id

    elem = Element("md:ReferenceValue", id=id_)

    if isinstance(obj, model.TargetReportPeriod):
        elem.append(Element("md:ReportPeriod", obj.report_period))
    elif isinstance(obj, model.TargetIdentifiableObject):
        elem.append(
            Element("md:ObjectReference", Element("URN", sdmx.urn.make(obj.obj)))
        )
    else:  # pragma: no cover
        raise NotImplementedError(type(obj))

    return elem


@writer
def _ra(obj: model.ReportedAttribute):
    child = []
    attrib: MutableMapping[str, str] = dict()

    if isinstance(obj.value_for, str):
        # NB value_for should be MetadataAttribute, but currently not due to limitations
        #    of .reader.xml.v21
        attrib.update(id=obj.value_for)
    else:  # pragma: no cover
        attrib.update(id=obj.value_for.id)

    if isinstance(obj, model.OtherNonEnumeratedAttributeValue):
        # Only write the "value" attribute if defined; some attributes are only
        # containers for child attributes
        if obj.value:
            attrib.update(value=obj.value)
    elif isinstance(obj, model.XHTMLAttributeValue):
        child.append(Element("com:StructuredText", obj.value))
    else:  # pragma: no cover
        raise NotImplementedError

    if len(obj.child):
        # Add child ReportedAttribute within an AttributeSet
        child.append(
            Element("md:AttributeSet", *[writer.recurse(ra) for ra in obj.child])
        )

    return Element("md:ReportedAttribute", *child, **attrib)



import json
import logging
import re
from io import BytesIO

import pandas as pd
import pytest

import sdmx
from sdmx.util.requests import save_response


def test_deprecated_request(caplog):
    message = "Request class will be removed in v3.0; use Client(…)"
    with pytest.warns(DeprecationWarning, match=re.escape(message)):
        sdmx.Request("ECB")

    assert caplog.record_tuples == [("sdmx.client", logging.WARNING, message)]


def test_read_sdmx(tmp_path, specimen):
    # Copy the file to a temporary file with an urecognizable suffix
    target = tmp_path / "foo.badsuffix"
    with specimen("flat.json", opened=False) as original:
        target.open("w").write(original.read_text())

    # With unknown file extension, read_sdmx() peeks at the file content
    sdmx.read_sdmx(target)

    # Format can be inferred from an already-open file without extension
    with specimen("flat.json") as f:
        sdmx.read_sdmx(f)

    # Exception raised when the file contents don't allow to guess the format
    bad_file = BytesIO(b"#! neither XML nor JSON")
    with pytest.raises(RuntimeError, match="cannot infer SDMX message format from "):
        sdmx.read_sdmx(bad_file)

    # Using the format= argument forces a certain reader to be used
    with pytest.raises(json.JSONDecodeError):
        sdmx.read_sdmx(bad_file, format="JSON")


class TestClient:
    @pytest.fixture
    def client(self, testsource, session_with_stored_responses):
        """A :class:`Client` connected to a non-existent test source."""
        return sdmx.Client(testsource, session=session_with_stored_responses)

    def test_init(self):
        with pytest.warns(
            DeprecationWarning, match=re.escape("Client(…, log_level=…) parameter")
        ):
            sdmx.Client(log_level=logging.ERROR)

        # Invalid source name raise an exception
        with pytest.raises(ValueError):
            sdmx.Client("noagency")

    # Regular methods
    def test_clear_cache(self, client):
        client.clear_cache()

    def test_session_attrs0(self, caplog, client):
        # Deprecated attributes
        with pytest.warns(DeprecationWarning, match="Setting Client.timeout"):
            client.timeout = 300

        with pytest.warns(DeprecationWarning, match="Getting Client.timeout"):
            assert client.timeout == 300

        client.get(
            "datastructure",
            dry_run=True,
            verify=True,  # Session attribute
            allow_redirects=False,  # Argument to Session.send()
        )

        assert not any("replaces" in m for m in caplog.messages)
        caplog.clear()

        # Same, with different values
        client.get(
            "datastructure",
            dry_run=True,
            verify=False,
            allow_redirects=True,
            timeout=123,
        )

        # Messages are logged
        assert "Client.session.verify=False replaces True" in caplog.messages
        assert (
            "Client.get() args {'allow_redirects': True, 'timeout': 123} replace "
            "{'allow_redirects': False}" in caplog.messages
        )

    def test_session_attrs1(self, testsource, session_with_stored_responses):
        with pytest.raises(ValueError):
            sdmx.Client(testsource, session=session_with_stored_responses, verify=False)

    def test_dir(self, client):
        """dir() includes convenience methods for resource endpoints."""
        expected = {
            "cache",
            "clear_cache",
            "get",
            "preview_data",
            "series_keys",
            "session",
            "source",
            "timeout",
        }
        expected |= set(ep.name for ep in sdmx.Resource)
        assert set(filter(lambda s: not s.startswith("_"), dir(client))) == expected

    def test_get0(self, client):
        """:meth:`.get` handles mixed query parameters correctly."""
        req = client.get(
            "dataflow", detail="full", params={"references": "none"}, dry_run=True
        )
        assert (
            "https://example.com/sdmx-rest/dataflow/TEST/all/latest?detail=full&"
            "references=none" == req.url
        )

    def test_get1(self, client):
        """Exceptions are raised on invalid arguments."""
        # Exception is raised on unrecognized arguments
        exc = "Unexpected/unhandled parameters {'foo': 'bar'}"
        with pytest.raises(ValueError, match=exc):
            client.get("datastructure", foo="bar")

    def test_getattr(self, client):
        with pytest.raises(AttributeError):
            client.notanendpoint()

    def test_request_from_args(self, caplog, client):
        # Raises for invalid resource type
        # TODO Move this test; this error is no longer handled in _request_from_args()
        kwargs = dict(resource_type="foo")
        with pytest.raises(AttributeError):
            client._request_from_args(kwargs)

        # Raises for not implemented endpoint
        _id = "OECD_JSON"
        with pytest.raises(NotImplementedError, match=f"{_id} does not implement"):
            sdmx.Client(_id).get("datastructure")

        # Raises for invalid key type
        with pytest.raises(TypeError, match="must be str or dict; got int"):
            client.get("data", key=12345)

        # Warns for deprecated argument
        with pytest.warns(
            DeprecationWarning, match="validate= keyword argument to Client.get"
        ):
            client.get("datastructure", validate=False, dry_run=True)

    # TODO update or remove
    @pytest.mark.xfail(reason="SDMX 3.0.0 is now supported → no exception raised")
    def test_v3_unsupported(self, client):
        """Client raises an exception when an SDMX 3.0 message is returned."""
        df_id, key = "DATAFLOW", ".KEY2.KEY3..KEY5"

        save_response(
            client.session,
            "GET",
            url=f"{client.source.url}/data/{df_id}/{key}",
            content="""<?xml version='1.0' encoding='UTF-8'?>
<mes:Structure xmlns:mes="http://www.sdmx.org/resources/sdmxml/schemas/v3_0/message">
</mes:Structure>""".encode(),
            headers={"Content-type": "application/vnd.sdmx.data+xml; version=3.0.0"},
        )

        with pytest.raises(
            ValueError, match="can't determine a reader for response content type"
        ):
            client.get("data", resource_id=df_id, key=key)


@pytest.mark.network
def test_request_get_args():
    ESTAT = sdmx.Client("ESTAT")

    # Client._make_key accepts '+'-separated values
    args = dict(
        resource_id="UNE_RT_A",
        key={"geo": "EL+ES+IE"},
        params={"startPeriod": "2007"},
        dry_run=True,
        use_cache=True,
    )
    # Store the URL
    url = ESTAT.data(**args).url

    # Using an iterable of key values gives the same URL
    args["key"] = {"geo": ["EL", "ES", "IE"]}
    assert ESTAT.data(**args).url == url

    # Using a direct string for a key gives the same URL
    args["key"] = "....EL+ES+IE"  # No specified values for first 4 dimensions
    assert ESTAT.data(**args).url == url

    # Giving 'provider' is redundant for a data request, causes a warning
    with pytest.warns(UserWarning, match="'agency_id' argument is redundant"):
        ESTAT.data(
            "UNE_RT_A",
            key={"geo": "EL+ES+IE"},
            params={"startPeriod": "2007"},
            agency_id="ESTAT",
        )

    # Using an unknown endpoint is an exception
    with pytest.raises(KeyError):
        ESTAT.get("badendpoint", "id")

    # TODO test Client.get(obj) with IdentifiableArtefact subclasses


@pytest.mark.network
def test_read_url0():
    """URL can be queried without instantiating Client."""
    sdmx.read_url(
        "https://sdw-wsrest.ecb.europa.eu/service/datastructure/ECB/ECB_EXR1/latest?"
        "references=all"
    )


def test_read_url1():
    """Exception is raised on invalid arguments."""
    with pytest.raises(
        ValueError, match=r"{'foo': 'bar'} supplied with get\(url=...\)"
    ):
        sdmx.read_url("https://example.com", foo="bar")


# @pytest.mark.skip(reason="Temporarily offline on 2021-03-23")
@pytest.mark.network
def test_request_preview_data():
    ECB = sdmx.Client("ECB")

    # List of keys can be retrieved
    keys = ECB.preview_data("EXR")
    assert isinstance(keys, list)

    # Count of keys can be determined
    assert len(keys) > 1000

    # A filter can be provided, resulting in fewer keys
    keys = ECB.preview_data("EXR", {"CURRENCY": "CAD+CHF+CNY"})
    assert len(keys) == 24

    # Result can be converted to pandas object
    keys_pd = sdmx.to_pandas(keys)
    assert isinstance(keys_pd, pd.DataFrame)
    assert len(keys_pd) == 24



import pytest

from sdmx import Resource, model
from sdmx.model import v21, v30

CLASSES = [
    # Appearing in .model.common
    "Annotation",
    "AnnotableArtefact",
    "IdentifiableArtefact",
    "NameableArtefact",
    "VersionableArtefact",
    "MaintainableArtefact",
    "ActionType",
    "ConstraintRoleType",
    "FacetValueType",
    "ExtendedFacetValueType",
    "Item",
    "ItemScheme",
    "FacetType",
    "Facet",
    "Representation",
    "Code",
    "Codelist",
    "ISOConceptReference",
    "Concept",
    "ConceptScheme",
    "Component",
    "ComponentList",
    "Category",
    "CategoryScheme",
    "Categorisation",
    "Contact",
    "Organisation",
    "Agency",
    "OrganisationScheme",
    "AgencyScheme",
    "Structure",
    "StructureUsage",
    "DimensionComponent",
    "Dimension",
    "TimeDimension",
    "DimensionDescriptor",
    "GroupDimensionDescriptor",
    "AttributeRelationship",
    "DimensionRelationship",
    "GroupRelationship",
    "DataAttribute",
    "AttributeDescriptor",
    "KeyValue",
    "TimeKeyValue",
    "AttributeValue",
    "Key",
    "GroupKey",
    "SeriesKey",
    "CodingFormat",
    "Level",
    "Hierarchy",
    "HierarchicalCode",
    "ConstraintRole",
    "ConstrainableArtefact",
    "SelectionValue",
    "MemberValue",
    "TimeRangeValue",
    "BeforePeriod",
    "AfterPeriod",
    "StartPeriod",
    "EndPeriod",
    "RangePeriod",
    "CubeRegion",
    "MetadataTargetRegion",
    "DataConsumer",
    "DataProvider",
    "DataConsumerScheme",
    "DataProviderScheme",
    "Datasource",
    "SimpleDatasource",
    "QueryDatasource",
    "RESTDatasource",
    "ProvisionAgreement",
    "CustomType",
    "CustomTypeScheme",
    "NamePersonalisation",
    "NamePersonalisationScheme",
    "Ruleset",
    "RulesetScheme",
    "Transformation",
    "UserDefinedOperator",
    "UserDefinedOperatorScheme",
    "FromVTLSpaceKey",
    "ToVTLSpaceKey",
    "VTLConceptMapping",
    "VTLDataflowMapping",
    "VTLMappingScheme",
    "TransformationScheme",
    # Appearing in model.InternationalString
    "DEFAULT_LOCALE",
    "InternationalString",
    # Appearing in model.Version
    "Version",
    # Classes that are distinct in .model.v21 versus .model.v30
    "SelectionValue",
    "MemberValue",
    "TimeRangeValue",
    "BeforePeriod",
    "AfterPeriod",
    "RangePeriod",
    "DataKey",
    "DataKeySet",
    "Constraint",
    "MemberSelection",
    "MeasureDescriptor",
    "DataStructureDefinition",
    "Observation",
    "DataSet",
    "StructureSpecificDataSet",
    "MetadataStructureDefinition",
    "TextAttributeValue",
    "XHTMLAttributeValue",
    "MetadataSet",
]

V21_ONLY = [
    "ContentConstraint",
    "PrimaryMeasure",
    "NoSpecifiedRelationship",
    "PrimaryMeasureRelationship",
    "ReportingYearStartDay",
    "MeasureDimension",
    "DataflowDefinition",
    "GenericDataSet",
    "GenericTimeSeriesDataSet",
    "StructureSpecificTimeSeriesDataSet",
    "ReportingCategory",
    "ReportingTaxonomy",
    "TargetObject",
    "DataSetTarget",
    "DimensionDescriptorValuesTarget",
    "IdentifiableObjectTarget",
    "ReportPeriodTarget",
    "MetadataTarget",
    "ReportStructure",
    "MetadataStructureDefinition",
    "MetadataflowDefinition",
    "TargetObjectValue",
    "TargetReportPeriod",
    "TargetIdentifiableObject",
    "TargetObjectKey",
    "ReportedAttribute",
    "EnumeratedAttributeValue",
    "NonEnumeratedAttributeValue",
    "OtherNonEnumeratedAttributeValue",
    "MetadataReport",
    "HierarchicalCodelist",
    "ItemAssociation",
    "CodeMap",
    "ItemSchemeMap",
    "CodelistMap",
    "StructureSet",
]

V30_ONLY = [
    "CodelistExtension",
    "GeoRefCode",
    "GeoGridCode",
    "GeoFeatureSetCode",
    "GeoCodelist",
    "GeographicCodelist",
    "GeoGridCodelist",
    "ValueItem",
    "ValueList",
    "MetadataProvider",
    "MetadataProviderScheme",
    "Measure",
    "Dataflow",  # Instead of DataflowDefinition
    "HierarchyAssociation",
    "DataflowRelationship",
    "MeasureRelationship",
    "ObservationRelationship",
    "DataConstraint",
    "MetadataConstraint",
    "MetadataAttributeDescriptor",
    "IdentifiableObjectSelection",
    "MetadataStructureDefinition",
    "Metadataflow",  # Instead of MetadataflowDefinition
    "MetadataAttributeValue",
    "CodedMetadataAttributeValue",
    "UncodedMetadataAttributeValue",
    "OtherUncodedAttributeValue",
    "TargetIdentifiableObject",
]


@pytest.mark.parametrize("module, extra", [(v21, V21_ONLY), (v30, V30_ONLY)])
def test_complete(module, extra):
    """:mod:`.model.v21` and :mod:`model.v30` each expose a complete set of classes."""
    # Each class is available using module.__getattr__
    for name in CLASSES:
        getattr(module, name)

    assert set(CLASSES + extra) == set(dir(module))


@pytest.mark.parametrize(
    "args, expected",
    [
        pytest.param(
            dict(name="Category", package="codelist"),
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, reason="Package 'codelist' invalid for Category"
            ),
        ),
        # Resource types appearing in StructureMessage
        (dict(name=Resource.agencyscheme), model.AgencyScheme),
        (dict(name=Resource.categorisation), model.Categorisation),
        (dict(name=Resource.categoryscheme), model.CategoryScheme),
        (dict(name=Resource.codelist), model.Codelist),
        (dict(name=Resource.conceptscheme), model.ConceptScheme),
        (dict(name=Resource.contentconstraint), v21.ContentConstraint),
        (dict(name=Resource.dataflow), v21.DataflowDefinition),
        (dict(name=Resource.metadatastructure), v21.MetadataStructureDefinition),
        (dict(name=Resource.organisationscheme), model.OrganisationScheme),
        (dict(name=Resource.provisionagreement), v21.ProvisionAgreement),
        pytest.param(
            dict(name=Resource.structure),
            v21.DataStructureDefinition,
            marks=pytest.mark.skip(reason="Ambiguous value, not implemented"),
        ),
    ],
)
def test_get_class_v21(args, expected) -> None:
    assert expected is model.v21.get_class(**args)


@pytest.mark.parametrize("args, expected", ((dict(name="ValueList"), v30.ValueList),))
def test_get_class_v30(args, expected) -> None:
    assert expected is model.v30.get_class(**args)


@pytest.mark.parametrize("klass, expected", (("ValueList", "codelist"),))
def test_package(klass, expected) -> None:
    assert expected == model.v30.PACKAGE[klass]


def test_deprecated_import():
    """Deprecation warning when importing SDMX 2.1-specific class from :mod:`.model`."""
    with pytest.warns(
        DeprecationWarning, match=r"DataStructureDefinition from sdmx\.model"
    ):
        model.DataStructureDefinition

    with pytest.raises(ImportError):
        from sdmx.model import Foo  # noqa: F401


def test_dir():
    """:func:`dir` gives only classes in :mod:`.model.common`."""
    assert "CategoryScheme" in dir(model)
    assert "DataStructureDefinition" not in dir(model)



"""Information related to the SDMX-REST web service standard."""

import abc
import re
from copy import copy
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any, ClassVar, Mapping, Optional
from urllib.parse import urlsplit, urlunsplit

if TYPE_CHECKING:
    import sdmx.source

# Mapping from Resource value to class name.
CLASS_NAME = {
    "dataflow": "DataflowDefinition",
    "datastructure": "DataStructureDefinition",
    "metadatastructure": "MetadataStructureDefinition",
}

# Inverse of :data:`CLASS_NAME`.
VALUE = {v: k for k, v in CLASS_NAME.items()}

#: Response codes defined by the SDMX-REST standard.
RESPONSE_CODE = {
    200: "OK",
    304: "No changes",
    400: "Bad syntax",
    401: "Unauthorized",
    403: "Semantic error",  # or "Forbidden"
    404: "Not found",
    406: "Not acceptable",
    413: "Request entity too large",
    414: "URI too long",
    500: "Internal server error",
    501: "Not implemented",
    503: "Unavailable",
}


class Resource(str, Enum):
    """Enumeration of SDMX-REST API resources.

    This class merges the "resources" defined in Section V of the SDMX 2.1 and 3.0
    standards; in the latter, only five values (availability, data, metadata, schema,
    structure) are used as the first part of a URL path; however, the choice of this
    first part and allowable query parameters depend on the more detailed list.

    ============================= ======================================================
    :class:`Enum` member          :mod:`sdmx.model` class
    ============================= ======================================================
    ``actualconstraint``          :class:`.ContentConstraint`
    ``agencyscheme``              :class:`.AgencyScheme`
    ``allowedconstraint``         :class:`.ContentConstraint`
    ``attachementconstraint``     :class:`.AttachmentConstraint`
    ``availableconstraint``       :class:`.ContentConstraint`
    ``categorisation``            :class:`.Categorisation`
    ``categoryscheme``            :class:`.CategoryScheme`
    ``codelist``                  :class:`.Codelist`
    ``conceptscheme``             :class:`.ConceptScheme`
    ``contentconstraint``         :class:`.ContentConstraint`
    ``customtypescheme``          :class:`.CustomTypeScheme`.
    ``data``                      :class:`.DataSet`
    ``dataflow``                  :class:`Dataflow(Definition) <.BaseDataflow>`
    ``dataconsumerscheme``        :class:`.DataConsumerScheme`
    ``dataproviderscheme``        :class:`.DataProviderScheme`
    ``datastructure``             :class:`DataStructureDefinition <.BaseDataStructureDefinition>`
    ``hierarchicalcodelist``      :class:`.v21.HierarchicalCodelist`.
    ``metadata``                  :class:`MetadataSet <.BaseMetadataSet>`.
    ``metadataflow``              :class:`Metadataflow(Definition) <.Metadataflow>`
    ``metadatastructure``         :class:`MetadataStructureDefinition <.BaseMetadataStructureDefinition>`
    ``namepersonalisationscheme`` :class:`.NamePersonalisationScheme`.
    ``organisationscheme``        :class:`.OrganisationScheme`
    ``provisionagreement``        :class:`.ProvisionAgreement`
    ``rulesetscheme``             :class:`.RulesetScheme`.
    ``structure``                 Mixed.
    ``structureset``              :class:`.StructureSet`.
    ``transformationscheme``      :class:`.TransformationScheme`.
    ``userdefinedoperatorscheme`` :class:`.UserdefinedoperatorScheme`.
    ``vtlmappingscheme``          :class:`.VTLMappingScheme`.
    ----------------------------- ------------------------------------------------------
    ``organisationunitscheme``    Not implemented.
    ``process``                   Not implemented.
    ``reportingtaxonomy``         Not implemented.
    ``schema``                    Not implemented.
    ============================= ======================================================

    """  # noqa: E501

    actualconstraint = "actualconstraint"
    agencyscheme = "agencyscheme"
    allowedconstraint = "allowedconstraint"
    attachementconstraint = "attachementconstraint"
    availableconstraint = "availableconstraint"
    categorisation = "categorisation"
    categoryscheme = "categoryscheme"
    codelist = "codelist"
    conceptscheme = "conceptscheme"
    contentconstraint = "contentconstraint"
    customtypescheme = "customtypescheme"
    data = "data"
    dataconsumerscheme = "dataconsumerscheme"
    dataflow = "dataflow"
    dataproviderscheme = "dataproviderscheme"
    datastructure = "datastructure"
    hierarchicalcodelist = "hierarchicalcodelist"
    metadata = "metadata"
    metadataflow = "metadataflow"
    metadatastructure = "metadatastructure"
    namepersonalisationscheme = "namepersonalisationscheme"
    organisationscheme = "organisationscheme"
    organisationunitscheme = "organisationunitscheme"
    process = "process"
    provisionagreement = "provisionagreement"
    registration = "registration"
    reportingtaxonomy = "reportingtaxonomy"
    rulesetscheme = "rulesetscheme"
    schema = "schema"
    structure = "structure"
    structureset = "structureset"
    transformationscheme = "transformationscheme"
    userdefinedoperatorscheme = "userdefinedoperatorscheme"
    vtlmappingscheme = "vtlmappingscheme"

    @classmethod
    def from_obj(cls, obj):
        """Return an enumeration value based on the class of `obj`."""
        value = obj.__class__.__name__
        return cls[VALUE.get(value, value)]

    @classmethod
    def class_name(cls, value: "Resource", default=None) -> str:
        """Return the name of a :mod:`sdmx.model` class from an enum value.

        Values are returned in lower case.
        """
        return CLASS_NAME.get(value.value, value.value)

    @classmethod
    def describe(cls):
        return "{" + " ".join(v.name for v in cls._member_map_.values()) + "}"


@dataclass
class Parameter(abc.ABC):
    """SDMX query parameter."""

    #: Keyword argument to :class:`.URL` understood by this parameter.
    name: str

    #: Allowable values.
    values: set = field(default_factory=set)

    #: Default value.
    default: Optional[str] = None

    @abc.abstractmethod
    def handle(self, parameters: dict[str, Any]) -> dict[str, str]:
        """Return a dict to update :attr:`.URL.path` or :attr:`.URL.query`."""


@dataclass
class PathParameter(Parameter):
    """SDMX query parameter appearing as a part of the path component of a URL."""

    def handle(self, parameters):
        """Return a length-1 dict to update :attr:`.URL.path`."""
        # Retrieve the value from `parameters`
        value = parameters.pop(self.name, None) or self.default

        # Check against allowable values
        if value is None:
            raise ValueError(f"Missing required parameter {self.name!r}")
        elif len(self.values) and value not in self.values:
            raise ValueError(f"{self.name}={value!r} not among {self.values}")

        # Return
        return {self.name: value}


@dataclass
class OptionalPath(Parameter):
    """Like :class:`.PathParameter`, but optional.

    If the corresponding keyword is missing
    """

    def handle(self, parameters):
        if value := parameters.pop(self.name, None):
            assert value in self.values or 0 == len(self.values)
            return {self.name: value}
        else:
            return {}


@dataclass
class QueryParameter(PathParameter):
    """SDMX query parameter appearing as part of the query string component of a URL.

    This version also responds to, for instance, :py:`fooBar="..."` when the
    :attr:`.name` is :py:`"foo_bar"`.
    """

    def __post_init__(self):
        # Convert self.name to lowerCamelCase as appearing in query strings
        self.camelName = re.sub(r"_([a-z])", lambda x: x.group(1).upper(), self.name)

    def handle(self, parameters):
        """Return a length-0 or -1 dict to update :attr:`.URL.query`."""
        if present := {self.name, self.camelName} & set(parameters):
            if 2 == len(present):
                raise ValueError(f"Cannot give both {self.name} and {self.camelName}")

            value = parameters.pop(present.pop())
            if value is None:
                return {}
            elif len(self.values) and value not in self.values:
                raise ValueError(f"{self.name}={value!r} not among {self.values}")

            return {self.camelName: value}
        else:
            return {}


@dataclass
class PositiveIntParam(QueryParameter):
    """A query parameter that must be a positive integer."""

    def handle(self, parameters):
        result = super().handle(parameters)
        try:
            k = list(result)[0]
        except IndexError:
            return result
        try:
            assert result[k] > 0
            result[k] = str(result[k])
        except (AssertionError, TypeError):
            raise ValueError(f"{k} must be positive integer; got {result[k]}")
        else:
            return result


# Todo: transcribe:
# - common:IDType
# - common:NCNameIDType
# - common:VersionType

PARAM: dict[str, Parameter] = {
    # Path parameters
    "agency_id": PathParameter("agency_id"),
    "key": OptionalPath("key"),
    "resource_id": PathParameter("resource_id", set(), "all"),
    #
    # Query parameters
    "dimension_at_observation": QueryParameter("dimension_at_observation"),
    "first_n_observations": PositiveIntParam("first_n_observations"),
    "include_history": QueryParameter("include_history", {True, False}),
    "last_n_observations": PositiveIntParam("last_n_observations"),
    "mode": QueryParameter("mode", {"available", "exact"}),
    "references_a": QueryParameter(
        "references",
        {
            "all",
            "codelist",
            "conceptscheme",
            "dataflow",
            "dataproviderscheme",
            "datastructure",
            "none",
        },
    ),
    "updated_after": QueryParameter("updated_after"),  # Also availability
}

#: Common lists of names used in both :data:`.v21.PARAM` and :data:`.v30.PARAM`. The
#: parameters in the latter generally have 1 or more additional entries.
NAMES = {
    "context": {
        "dataflow",
        "datastructure",
        "metadataflow",
        "metadatastructure",
        "provisionagreement",
    },
    "detail_s": {
        "allcompletestubs",
        "allstubs",
        "full",
        "referencecompletestubs",
        "referencepartial",
        "referencestubs",
    },
    "references_s": {
        "all",
        "children",
        "descendants",
        "none",
        "parents",
        "parentsandsiblings",
    },
}


class URL(abc.ABC):
    """Abstract base utility class to build SDMX-REST API URLs.

    Parameters
    ----------
    source : .Source
        Provides a base URL (API entry point) and optional modification of the complete
        URL.
    resource_type : .Resource
        Indicates the type of query to be made.
    """

    #: SDMX REST web service to query.
    source: "sdmx.source.Source"

    #: Type of resource to be retrieved; a member of :class:`.Resource`.
    resource_type: Resource

    #: Pieces for the hierarchical path component of the URL. If
    _path: dict[str, Optional[str]]

    #: Pieces for the query component of the URL.
    query: dict[str, str]

    # Keyword arguments to the constructor
    _params: dict

    _all_parameters: ClassVar[Mapping[Any, Parameter]]

    def __init__(self, source: "sdmx.source.Source", resource_type: Resource, **kwargs):
        # Check for duplicates in kwargs["params"] and kwargs directly
        params = copy(kwargs)
        params_dict = params.pop("params", {})
        if overlap := set(params) & set(params_dict):
            raise ValueError(f"Duplicate values for query parameters {overlap}")
        params.update(params_dict)

        self.source = source
        self.resource_type = resource_type

        # Store the keyword arguments
        self._params = params

        # Internal
        self._path = dict()
        self.query = dict()

        # Identify the query type
        if resource_type.name in {
            "data",
            "metadata",
            "schema",
            "structure",
            "registration",
        }:
            query_type = resource_type.name
        elif resource_type.name in {"availableconstraint"}:
            query_type = "availability"
        else:
            query_type = "structure"

        # Dispatch to a method appropriate to the Version and QueryType
        getattr(self, f"handle_{query_type}")()

        if len(self._params):
            raise ValueError(f"Unexpected/unhandled parameters {self._params}")

    # General-purpose methods

    def handle_path_params(self, expr: str) -> None:
        """Extend :attr:`.path` with parts from `expr`, a "/"-delimited string.

        In an `expr` such as :py:`"/foo/{param}/"`:

        - Parts enclosed in braces like :py:`"{param}"` are looked up in
          :attr:`._all_parameters``, and the resulting :class:`.Parameter` instance
          consumes a keyword argument.
        - Parts that are plain strings like :py:`"foo"` are passed through directly.
        """
        for part in expr.split("/"):
            if part.startswith("{"):
                param = self._all_parameters[part.strip("{}")]
                self._path.update(param.handle(self._params))
            else:
                self._path.update({part: None})

    def handle_query_params(self, expr: str) -> None:
        """Extend :attr:`.query` with parts from `expr`, a " "-delimited string."""
        for p in map(self._all_parameters.__getitem__, expr.split()):
            self.query.update(p.handle(self._params))

    # Handlers for different QueryTypes
    @abc.abstractmethod
    def handle_availability(self) -> None:
        """Handle URL parameters for availability endpoints (abstract method)."""

    @abc.abstractmethod
    def handle_data(self) -> None:
        """Handle URL parameters for data endpoints (abstract method)."""

    @abc.abstractmethod
    def handle_metadata(self) -> None:
        """Handle URL parameters for metadata endpoints (abstract method)."""

    @abc.abstractmethod
    def handle_registration(self) -> None:
        """Handle URL parameters for registration endpoints (abstract method)."""

    def handle_schema(self) -> None:
        """Handle URL parameters for schema endpoints."""
        self._params.setdefault("agency_id", self.source.id)
        self.handle_path_params(
            self.rt + "/{context}/{agency_id}/{resource_id}/{version}"
        )
        self.handle_query_params("dimension_at_observation")

    def handle_structure(self) -> None:
        """Handle URL parameters for structure endpoints."""
        self._params.setdefault("agency_id", self.source.id)
        self.handle_path_params("{agency_id}/{resource_id}/{version}")
        self.handle_query_params("detail_s references_s")

    def join(self, *, with_query: bool = True) -> str:
        """Join the URL parts, returning a complete URL.

        Parameters
        ----------
        with_query : bool
            If :any:`False`, omit :attr:`.query` from the joined URL.
        """
        # Keep the URL scheme, netloc, and any path from the source's base URL
        parts = list(urlsplit(self.source.url)[:3]) + [None, None]

        # Assemble path string
        parts[2] = re.sub("([^/])$", r"\1/", parts[2] or "") + "/".join(
            (value or name) for name, value in self._path.items()
        )

        if with_query:
            # Assemble query string
            parts[3] = "&".join(f"{k}={v}" for k, v in self.query.items())

        return urlunsplit(parts)

    @property
    def rt(self) -> str:
        """Shorthand access to :class:`str` name of :attr:`.resource_type`."""
        return self.resource_type.name



import re
from abc import abstractmethod
from collections import ChainMap, defaultdict
from importlib import import_module
from itertools import chain, count
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
)

from lxml import etree
from lxml.etree import QName

import sdmx.urn
from sdmx import message
from sdmx.exceptions import XMLParseError  # noqa: F401
from sdmx.format import Version, list_media_types
from sdmx.model import common
from sdmx.reader.base import BaseReader

if TYPE_CHECKING:
    import types

# Sentinel value for a missing Agency
_NO_AGENCY = common.Agency()


class NotReference(Exception):
    """Raised when the `elem` passed to :class:`.Reference` is not a reference."""


class BaseReference:
    """Temporary class for references.

    - `cls`, `id`, `version`, and `agency_id` are always for a MaintainableArtefact.
    - If the reference target is a MaintainableArtefact (`maintainable` is True),
      `target_cls` and `target_id` are identical to `cls` and `id`, respectively.
    - If the target is not maintainable, `target_cls` and `target_id` describe it.

    `cls_hint` is an optional hint for when the object is instantiated, i.e. a more
    specific override for `cls`/`target_cls`.
    """

    __slots__ = (
        "agency",
        "cls",
        "id",
        "maintainable",
        "target_cls",
        "target_id",
        "version",
    )

    def __init__(self, reader, elem, cls_hint=None):
        parent_tag = elem.tag

        info = self.info_from_element(elem)

        # Find the target class
        target_cls = reader.model.get_class(info["class"], info["package"])

        if target_cls is None:
            # Try the parent tag name
            target_cls = reader.format.class_for_tag(parent_tag)

        if cls_hint and (target_cls is None or issubclass(cls_hint, target_cls)):
            # Hinted class is more specific than target_cls, or failed to find a target
            # class above
            target_cls = cls_hint

        if target_cls is None:
            raise ValueError(f"Unable to determine target class for {info}", info)

        self.maintainable = issubclass(target_cls, common.MaintainableArtefact)

        if self.maintainable:
            # MaintainableArtefact is the same as the target
            cls, info["id"] = target_cls, info["target_id"]
        else:
            # Get the class for the parent MaintainableArtefact
            cls = reader.model.parent_class(target_cls)

        # Store
        self.cls = cls
        self.agency = (
            common.Agency(id=info["agency"]) if info.get("agency", None) else _NO_AGENCY
        )
        self.id = info["id"]
        self.version = info.get("version", None)
        self.target_cls = target_cls
        self.target_id = info["target_id"]

    @classmethod
    @abstractmethod
    def info_from_element(cls, elem) -> dict[str, Any]: ...

    def __str__(self):
        # NB for debugging only
        return (  # pragma: no cover
            f"{self.cls.__name__}={self.agency.id}:{self.id}({self.version}) → "
            f"{self.target_cls.__name__}={self.target_id}"
        )


class XMLEventReader(BaseReader):
    """Populate the parser, format, and model attributes of :class:`Reader`."""

    # Defined on BaseReader
    suffixes = [".xml"]

    #: SDMX-ML version handled by this reader.
    xml_version: ClassVar[Version]

    #: Reference to the module defining the format read.
    format: ClassVar["types.ModuleType"]
    #: Reference to the module defining the information model read.
    model: ClassVar["types.ModuleType"]

    #: :class:`.BaseReference` subclass used by this reader.
    Reference: ClassVar[type[BaseReference]]

    #: Mapping from (QName, ["start", "end"]) to a function that parses the
    #: element/event or else None (no parsing).
    parser: ClassVar[dict[tuple[QName, str], Callable]]

    # One-way counter for use in stacks
    _count: Iterator[int]

    def __init_subclass__(cls: type["XMLEventReader"]):
        # Empty dictionary
        cls.parser = {}

        name = {Version["2.1"]: "v21", Version["3.0.0"]: "v30"}[cls.xml_version]
        cls.format = import_module(f"sdmx.format.xml.{name}")
        cls.model = import_module(f"sdmx.model.{name}")
        cls.media_types = list_media_types(base="xml", version=cls.xml_version)

    def __init__(self):
        # Initialize counter
        self._count = count()

    # BaseReader methods

    def convert(
        self,
        data,
        structure=None,
        _events=None,
        **kwargs,
    ) -> message.Message:
        # Initialize stacks
        self.stack: dict[Union[type, str], dict[Union[str, int], Any]] = defaultdict(
            dict
        )

        # Elements to ignore when parsing finishes
        self.ignore = set()

        # If calling code provided a {Metad,D}ataStructureDefinition, add it to a stack,
        # and let it be ignored when parsing finishes
        structure = self._handle_deprecated_kwarg(structure, kwargs)
        self.push(structure)
        self.ignore.add(id(structure))

        if _events is None:
            events = cast(
                Iterator[tuple[str, etree._Element]],
                etree.iterparse(data, events=("start", "end")),
            )
        else:
            events = _events

        try:
            # Use the etree event-driven parser
            # NB (typing) iterparse() returns tuples. For "start" and "end", the second
            #    item is etree._Element, but for other events, e.g. "start-ns", it is
            #    not. types-lxml accurately reflects this. Narrow the type here for the
            #    following code.
            for event, element in events:
                try:
                    # Retrieve the parsing function for this element & event
                    func = self.parser[self.format.qname(element.tag), event]
                except KeyError:  # pragma: no cover
                    if QName(element.tag).namespace == "http://www.w3.org/1999/xhtml":
                        continue
                    # Don't know what to do for this (element, event)
                    raise NotImplementedError(element.tag, event) from None

                if func is None:
                    continue  # Explicitly no parser for this (element, event) → skip

                result = func(self, element)  # Parse the element
                self.push(result)  # Store the result

                if event == "end":
                    element.clear()  # Free memory

        except Exception as exc:
            # Parsing failed; display some diagnostic information
            self._dump()
            print(etree.tostring(element, pretty_print=True).decode())
            raise XMLParseError from exc

        # Parsing complete; count uncollected items from the stacks, which represent
        # parsing errors

        # Remove some internal items
        self.pop_single("SS without structure")
        self.pop_single("DataSetClass")

        # Count only non-ignored items
        uncollected = -1
        for key, objects in self.stack.items():
            uncollected += sum(
                [1 if id(o) not in self.ignore else 0 for o in objects.values()]
            )

        if uncollected > 0:  # pragma: no cover
            self._dump()
            raise RuntimeError(f"{uncollected} uncollected items")

        return cast(message.Message, self.get_single(message.Message, subclass=True))

    @classmethod
    def start(cls, names: str, only: bool = True):
        """Decorator for a function that parses "start" events for XML elements."""

        def decorator(func):
            for tag in map(cls.format.qname, names.split()):
                cls.parser[tag, "start"] = func
                if only:
                    cls.parser[tag, "end"] = None
            return func

        return decorator

    @classmethod
    def end(cls, names: str, only: bool = True):
        """Decorator for a function that parses "end" events for XML elements."""

        def decorator(func):
            for tag in map(cls.format.qname, names.split()):
                cls.parser[tag, "end"] = func
                if only:
                    cls.parser[tag, "start"] = None
            return func

        return decorator

    @classmethod
    def possible_reference(cls, cls_hint: Optional[type] = None, unstash: bool = False):
        """Decorator for a function where the `elem` parsed may be a Reference.

        Before calling the decorated function, attempt to parse the `elem` as a
        :class:`.Reference`. If successful, return the reference instead of calling the
        function. If `elem` does not contain a reference, call the decorated function.

        Parameters
        ----------
        cls_hint :
            Passed to :class:`.Reference`.
        unstash : bool, optional
            If :data:`True`, call :meth:`.unstash` after successfully resolving a
            reference.
        """

        def decorator(func):
            def wrapped(reader: "XMLEventReader", elem):
                try:
                    # Identify a reference
                    result = reader.Reference(
                        reader,
                        elem,
                        cls_hint=cls_hint or reader.class_for_tag(elem.tag),
                    )
                except NotReference:
                    # Call the wrapped function
                    result = func(reader, elem)
                else:
                    # Successful; unstash if configured
                    if unstash:
                        reader.unstash()

                return result

            return wrapped

        return decorator

    # Stack handling

    def _clean(self):  # pragma: no cover
        """Remove empty stacks."""
        for key in list(self.stack.keys()):
            if len(self.stack[key]) == 0:
                self.stack.pop(key)

    def _dump(self):  # pragma: no cover
        """Print the stacks, for debugging."""
        self._clean()
        print("\n\n")
        for key, values in self.stack.items():
            print(f"--- {key} ---")
            if isinstance(values, Mapping):
                print(
                    *map(lambda kv: f"{kv[0]} ({id(kv[1])}) {kv[1]!s}", values.items()),
                    sep="\n",
                    end="\n\n",
                )
        print("\nIgnore:\n", self.ignore)

    def push(self, stack_or_obj, obj=None):
        """Push an object onto a stack."""
        if stack_or_obj is None:
            return
        elif obj is None:
            # Add the object to a stack based on its class
            obj = stack_or_obj
            s = stack_or_obj.__class__
        elif isinstance(stack_or_obj, str):
            # Stack with a string name
            s = stack_or_obj
        else:
            # Element; use its local name
            s = QName(stack_or_obj).localname

        # Get the ID for the element in the stack: its .id attribute, if any, else a
        # unique number
        id = getattr(obj, "id", next(self._count)) or next(self._count)

        if id in self.stack[s]:
            # Avoid a collision for two distinct objects with the same ID, e.g. with
            # different maintainers (ECB:AGENCIES vs. SDMX:AGENCIES). Re-insert with
            # numerical keys. This means the objects cannot be retrieved by their ID,
            # but the code does not rely on this.
            self.stack[s][next(self._count)] = self.stack[s].pop(id)
            id = next(self._count)

        self.stack[s][id] = obj

    def stash(self, *stacks):
        """Temporarily hide all objects in the given `stacks`."""
        self.push("_stash", {s: self.stack.pop(s, dict()) for s in stacks})

    def unstash(self):
        """Restore the objects hidden by the last :meth:`stash` call to their stacks.

        Calls to :meth:`.stash` and :meth:`.unstash` should be matched 1-to-1; if the
        latter outnumber the former, this will raise :class:`.KeyError`.
        """
        for s, values in (self.pop_single("_stash") or {}).items():
            self.stack[s].update(values)

    # Delegate to version-specific module
    @classmethod
    def class_for_tag(cls, tag: str) -> type:
        return cls.format.class_for_tag(tag)

    @classmethod
    def qname(cls, ns_or_name, name=None) -> QName:
        return cls.format.qname(ns_or_name, name)

    def get_single(
        self,
        cls_or_name: Union[type, str],
        id: Optional[str] = None,
        version: Optional[str] = None,
        subclass: bool = False,
    ) -> Optional[Any]:
        """Return a reference to an object while leaving it in its stack.

        Always returns 1 object. Returns :obj:`None` if no matching object exists, or if
        2 or more objects meet the conditions.

        If `id` (and `version`) is/are given, only return an IdentifiableArtefact with
        the matching ID (and version).

        If `cls_or_name` is a class and `subclass` is :obj:`True`; check all objects in
        the stack `cls_or_name` *or any stack for a subclass of this class*.
        """
        if subclass:
            keys: Iterable[Union[type, str]] = filter(
                matching_class(cls_or_name), self.stack.keys()
            )
            results: Mapping = ChainMap(*[self.stack[k] for k in keys])
        else:
            results = self.stack.get(cls_or_name, dict())

        if id and version:
            for v in results.values():
                if v.id == id and v.version == version:
                    return v
            return None
        elif id:
            return results.get(id)
        elif len(results) != 1:
            # 0 or ≥2 results
            return None
        else:
            return next(iter(results.values()))

    def pop_all(self, cls_or_name: Union[type, str], subclass=False) -> Sequence:
        """Pop all objects from stack *cls_or_name* and return.

        If `cls_or_name` is a class and `subclass` is :obj:`True`; return all objects in
        the stack `cls_or_name` *or any stack for a subclass of this class*.
        """
        if subclass:
            keys: Iterable[Union[type, str]] = list(
                filter(matching_class(cls_or_name), self.stack.keys())
            )
            result: Iterable = chain(*[self.stack.pop(k).values() for k in keys])
        else:
            result = self.stack.pop(cls_or_name, dict()).values()

        return list(result)

    def pop_single(self, cls_or_name: Union[type, str]):
        """Pop a single object from the stack for `cls_or_name` and return."""
        try:
            return self.stack[cls_or_name].popitem()[1]
        except KeyError:
            return None

    def peek(self, cls_or_name: Union[type, str]):
        """Get the object at the top of stack `cls_or_name` without removing it."""
        try:
            key, value = self.stack[cls_or_name].popitem()
            self.stack[cls_or_name][key] = value
            return value
        except KeyError:  # pragma: no cover
            return None

    def pop_resolved_ref(self, cls_or_name: Union[type, str]):
        """Pop a reference to `cls_or_name` and resolve it."""
        return self.resolve(self.pop_single(cls_or_name))

    def reference(self, elem, cls_hint=None) -> "BaseReference":
        return self.Reference(self, elem, cls_hint=cls_hint)

    def resolve(self, ref):
        """Resolve the Reference instance `ref`, returning the referred object."""
        if not isinstance(ref, BaseReference):
            # None, already resolved, or not a Reference
            return ref

        # Try to get the target directly
        target = self.get_single(
            ref.target_cls, ref.target_id, ref.version, subclass=True
        )

        if target:
            return target

        # MaintainableArtefact with is_external_reference=True; either a new object, or
        # reference to an existing object
        target_or_parent = self.maintainable(
            ref.cls, None, id=ref.id, maintainer=ref.agency, version=ref.version
        )

        if ref.maintainable:
            # `target_or_parent` is the target
            return target_or_parent

        # At this point, trying to resolve a reference to a child object of a parent
        # MaintainableArtefact; `target_or_parent` is the parent
        parent = target_or_parent

        if parent.is_external_reference:
            # Create the child
            return parent.setdefault(id=ref.target_id)
        else:
            try:
                # Access the child. Mismatch here will raise KeyError
                return parent[ref.target_id]
            except KeyError:
                if isinstance(parent, common.ItemScheme):
                    return parent.get_hierarchical(ref.target_id)
                raise  # pragma: no cover

    def annotable(self, cls, elem, **kwargs):
        """Create a AnnotableArtefact of `cls` from `elem` and `kwargs`.

        Collects all parsed <com:Annotation>.
        """
        if elem is not None:
            kwargs.setdefault("annotations", [])
            kwargs["annotations"].extend(self.pop_all(common.Annotation))
        return cls(**kwargs)

    def identifiable(self, cls, elem, **kwargs):
        """Create a IdentifiableArtefact of `cls` from `elem` and `kwargs`."""
        setdefault_attrib(kwargs, elem, "id", "urn", "uri")
        return self.annotable(cls, elem, **kwargs)

    def nameable(self, cls, elem, **kwargs):
        """Create a NameableArtefact of `cls` from `elem` and `kwargs`.

        Collects all parsed :class:`.InternationalString` localizations of <com:Name>
        and <com:Description>.
        """
        obj = self.identifiable(cls, elem, **kwargs)
        if elem is not None:
            add_localizations(obj.name, self.pop_all("Name"))
            add_localizations(obj.description, self.pop_all("Description"))
        return obj

    def maintainable(self, cls, elem, **kwargs):
        """Create or retrieve a MaintainableArtefact of `cls` from `elem` and `kwargs`.

        Following the SDMX-IM class hierarchy, :meth:`maintainable` calls
        :meth:`nameable`, which in turn calls :meth:`identifiable`, etc. (Since no
        concrete class is versionable but not maintainable, no separate method is
        created, for better performance). For all of these methods:

        - Already-parsed items are removed from the stack only if `elem` is not
          :obj:`None`.
        - `kwargs` (e.g. 'id') take precedence over any values retrieved from
          attributes of `elem`.

        If `elem` is None, :meth:`maintainable` returns a MaintainableArtefact with
        the is_external_reference attribute set to :obj:`True`. Subsequent calls with
        the same object ID will return references to the same object.
        """
        setdefault_attrib(
            kwargs,
            elem,
            "isExternalReference",
            "isFinal",
            "validFrom",
            "validTo",
            "version",
        )

        # Ensure is_external_reference and is_final are bool
        try:
            kwargs["is_external_reference"] = kwargs["is_external_reference"] == "true"
        except (KeyError, SyntaxError):
            kwargs.setdefault("is_external_reference", elem is None)
        kwargs["is_final"] = kwargs.get("is_final", None) == "true"

        # Create a candidate object
        obj = self.nameable(cls, elem, **kwargs)

        try:
            # Retrieve the Agency.id for obj.maintainer
            maint = self.get_single(common.Agency, elem.attrib["agencyID"])
        except (AttributeError, KeyError):
            pass
        else:
            # Elem contains a maintainer ID
            if maint is None:
                # …but it did not correspond to an existing object; create one
                maint = common.Agency(id=elem.attrib["agencyID"])
                self.push(maint)
                # This object is never collected; ignore it at end of parsing
                self.ignore.add(id(maint))
            obj.maintainer = maint

        # Maybe retrieve an existing object of the same class, ID, and version (if any)
        existing = self.get_single(cls, obj.id, version=obj.version)

        if existing and (
            existing.compare(obj, strict=True)
            or (existing.urn or sdmx.urn.make(existing)) == sdmx.urn.make(obj)
        ):
            if elem is not None:
                # Update `existing` from `obj` to preserve references
                # If `existing` was a forward reference <Ref/>, its URN was not stored.
                for attr in list(kwargs.keys()) + ["urn"]:
                    setattr(existing, attr, getattr(obj, attr))

            # Discard the candidate
            obj = existing
        elif obj.is_external_reference:
            # A new external reference. Ensure it has a URN.
            obj.urn = obj.urn or sdmx.urn.make(obj)
            # Push onto the stack to be located by next calls
            self.push(obj)

        return obj


def add_localizations(target: common.InternationalString, values: list) -> None:
    """Add localized strings from *values* to *target*."""
    target.localizations.update({locale: label for locale, label in values})


def matching_class(cls):
    """Filter condition; see :meth:`.get_single` and :meth:`.pop_all`."""
    return lambda item: isinstance(item, type) and issubclass(item, cls)


def setdefault_attrib(target, elem, *names):
    """Update `target` from :py:`elem.attrib` for the given `names`."""
    try:
        for name in names:
            try:
                target.setdefault(to_snake(name), elem.attrib[name])
            except KeyError:
                pass
    except AttributeError:
        pass  # No elem.attrib; elem is None


TO_SNAKE_RE = re.compile("([A-Z]+)")


def to_snake(value):
    """Convert *value* from lowerCamelCase to snake_case."""
    return TO_SNAKE_RE.sub(r"_\1", value).lower()



import pytest

from sdmx.model.common import ConstraintRole, ConstraintRoleType
from sdmx.model.v30 import (
    DataConstraint,
    GeoCodelistType,
    GeoFeatureSetCode,
    GeographicCodelist,
    GeoGridCode,
    GeoGridCodelist,
    HierarchicalCode,
    Hierarchy,
    HierarchyAssociation,
    Level,
    MetadataConstraint,
    MetadataProvider,
    MetadataProviderScheme,
)

# §4.3: Codelist


class TestGeoGridCode:
    def test_init(self):
        GeoGridCode(geo_cell="foo")


class TestGeoFeatureSetCode:
    def test_init(self):
        GeoFeatureSetCode(value="foo")


class TestGeographicCodelist:
    def test_init(self):
        cl = GeographicCodelist()

        assert GeoCodelistType.geographic == cl.geo_type


class TestGeoGridCodelist:
    def test_init(self):
        cl = GeoGridCodelist()

        assert GeoCodelistType.geogrid == cl.geo_type


# §4.7: OrganisationScheme


class TestMetadataProvider:
    def test_init(self):
        MetadataProvider()


class TestMetadataProviderScheme:
    def test_init(self):
        MetadataProviderScheme()


# §8: Hierarchy


class TestLevel:
    def test_init(self):
        Level()


class TestHierarchicalCode:
    def test_init(self):
        HierarchicalCode()


class TestHierarchy:
    def test_init(self):
        Hierarchy()


class TestHierarchyAssociation:
    def test_init(self):
        HierarchyAssociation()


# §12.3: Constraints


_ROLE_PARAMS = [
    ConstraintRole(role=ConstraintRoleType.actual),
    ConstraintRole(role=ConstraintRoleType.allowable),
    ConstraintRoleType.actual,
    ConstraintRoleType.allowable,
    "actual",
    "allowable",
    pytest.param("foo", marks=pytest.mark.xfail(raises=KeyError)),
]


class TestDataConstraint:
    @pytest.mark.parametrize("role", _ROLE_PARAMS)
    def test_init(self, role):
        DataConstraint(role=role)


class TestMetadataConstraint:
    @pytest.mark.parametrize("role", _ROLE_PARAMS)
    def test_init(self, role):
        MetadataConstraint(role=role)



"""SDMX-REST API v1.5.0.

Note that version 1.5.0 of the REST API corresponds to version 2.1 of the overall SDMX
standards. See the `documentation
<https://github.com/sdmx-twg/sdmx-rest/tree/v1.5.0/v2_1/ws/rest/docs>`_ and `OpenAPI
specification
<https://github.com/sdmx-twg/sdmx-rest/blob/v1.5.0/v2_1/ws/rest/src/sdmx-rest.yaml>`_
for further details.
"""

from collections import ChainMap
from warnings import warn

from . import common
from .common import OptionalPath, PathParameter, QueryParameter, Resource

#: v1.5.0-specific path and query parameters.
PARAM: dict[str, common.Parameter] = {
    # Path parameters
    # NB the text and YAML OpenAPI specification disagree on whether this is required
    "component_id": OptionalPath("component_id"),
    "context": PathParameter("context", common.NAMES["context"]),
    "flow": OptionalPath("resource_id"),
    "provider": OptionalPath("provider"),
    "version": PathParameter("version", set(), "latest"),
    #
    # Query parameters
    "detail_d": QueryParameter(
        "detail", {"dataonly", "full", "nodata", "serieskeysonly"}
    ),
    "detail_s": QueryParameter("detail", common.NAMES["detail_s"]),
    "references_s": QueryParameter(
        "references", common.NAMES["references_s"] | set(Resource)
    ),
    "start_period": QueryParameter("start_period"),
    "end_period": QueryParameter("end_period"),
    "explicit_measure": QueryParameter("explicit_measure", {True, False}),
}


class URL(common.URL):
    """Utility class to build SDMX-REST API v1.5.0 URLs."""

    _all_parameters = ChainMap(common.PARAM, PARAM)

    def handle_availability(self):
        """Handle URL parameters for availability endpoints."""
        self.handle_path_params(self.rt + "/{flow}/{key}/{provider}/{component_id}")
        self.handle_query_params(
            "start_period end_period updated_after references_a mode"
        )

    def handle_data(self):
        """Handle URL parameters for data endpoints."""
        if self._params.pop("agency_id", None):
            warn("'agency_id' argument is redundant for data queries", UserWarning, 2)

        self.handle_path_params(self.rt + "/{flow}/{key}/{provider}")
        self.handle_query_params(
            "start_period end_period updated_after first_n_observations "
            "last_n_observations dimension_at_observation detail_d include_history"
        )

    def handle_metadata(self):
        """Handle URL parameters for metadata endpoints.

        Raises
        ------
        NotImplementedError
            Although it is described in the standards, there are no known SDMX REST data
            sources that support this API endpoint to confirm behaviour. If you are
            aware of one, please `open an issue
            <https://github.com/khaeru/sdmx/issues/new>`_.
        """
        raise NotImplementedError

    def handle_registration(self):
        """Handle URL parameters for registration endpoints.

        Raises
        ------
        ValueError
            Always. This type of query is not included in SDMX-REST v1.5.0.
        """
        raise ValueError(
            "/registration/… queries not supported in the SDMX-REST v1.5.0 API"
        )

    def handle_schema(self):
        """Handle URL parameters for schema endpoints."""
        super().handle_schema()
        self.handle_query_params("explicit_measure")

    def handle_structure(self):
        """Handle URL parameters for structure endpoints.

        .. warning::

           This method currently preserves the behaviour of :meth:`.Client.get` as of
           :mod:`sdmx` version 2.13.1 and earlier. Namely, defaults are supplied for the
           ``?references=…`` query parameter in the following cases:

           - :attr:`.Resource.dataflow` or :attr:`.Resource.datastructure` **and** a
             specific ``resource_id`` is given (that is, not "all") → default
             ``?references=all``.
           - :attr:`.Resource.categoryscheme` → default
             ``?references=parentsandsiblings``.

           These differ from the SDMX-REST v1.5.0 standard, which states the default
           should be ``none`` in all cases. The :mod:`sdmx`-specific defaults **may** be
           deprecated and removed in future versions; to ensure stable behaviour, give
           the parameter explicitly.
        """
        self.handle_path_params(self.rt)
        super().handle_structure()

        # Moved from Client._request_from_args()
        # TODO Consider deprecating and eventually removing these defaults
        if (
            self.resource_type in {Resource.dataflow, Resource.datastructure}
            and self._path["resource_id"] != "all"
        ):
            self.query.setdefault("references", "all")
        elif self.resource_type in {Resource.categoryscheme}:
            self.query.setdefault("references", "parentsandsiblings")
